import sys
import numpy, matplotlib

from silx.gui.plot import Plot2D

from silx.gui.plot.StackView import StackViewMainWindow

from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import QRect
from PyQt5.QtWidgets import QApplication

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas

from srxraylib.plot import gol

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import OWAction
from oasys.widgets import widget
from oasys.widgets import gui as oasysgui
from oasys.widgets.exchange import DataExchangeObject
from oasys.util.oasys_util import EmittingStream

class ESRFWidget(widget.OWWidget):
    author = "Luca Rebuffi"
    maintainer_email = "lrebuffi@anl.gov"

    outputs = [{"name": "esrf_data",
                "type": DataExchangeObject,
                "doc": ""}]

    IMAGE_WIDTH = 760
    IMAGE_HEIGHT = 545
    MAX_WIDTH = 1320
    MAX_HEIGHT = 700
    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 560

    view_type=Setting(1)

    calculated_data = None

    want_main_area = 1

    def __init__(self):
        super().__init__()

        self.runaction = OWAction("Compute", self)
        self.runaction.triggered.connect(self.compute)
        self.addAction(self.runaction)

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.MAX_WIDTH)),
                               round(min(geom.height()*0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        box0 = gui.widgetBox(self.controlArea, "", orientation="horizontal")
        #widget buttons: compute, set defaults, help
        gui.button(box0, self, "Compute", callback=self.compute)
        gui.button(box0, self, "Defaults", callback=self.defaults)
        gui.button(box0, self, "Help", callback=self.help1)

        gui.separator(self.controlArea, height=10)

        self.build_gui()

        gui.rubber(self.controlArea)

        self.main_tabs = oasysgui.tabWidget(self.mainArea)
        plot_tab = oasysgui.createTabPage(self.main_tabs, "Results")
        out_tab = oasysgui.createTabPage(self.main_tabs, "Output")

        view_box = oasysgui.widgetBox(plot_tab, "Results Options", addSpace=False, orientation="horizontal")
        view_box_1 = oasysgui.widgetBox(view_box, "", addSpace=False, orientation="vertical", width=350)

        self.view_type_combo = gui.comboBox(view_box_1, self, "view_type", label="View Results",
                                            labelWidth=220,
                                            items=["No", "Yes"],
                                            callback=self.set_ViewType, sendSelectedValue=False, orientation="horizontal")

        self.tab = []
        self.tabs = oasysgui.tabWidget(plot_tab)

        self.initializeTabs()

        self.esrf_output = QtWidgets.QTextEdit()
        self.esrf_output.setReadOnly(True)

        out_box = gui.widgetBox(out_tab, "System Output", addSpace=True, orientation="horizontal")
        out_box.layout().addWidget(self.esrf_output)

        self.esrf_output.setFixedHeight(600)
        self.esrf_output.setFixedWidth(600)

        self.current_tab = -1

        gui.rubber(self.mainArea)

    def build_gui(self):
        pass

    def initializeTabs(self):
        size = len(self.tab)
        indexes = range(0, size)

        for index in indexes:
            self.tabs.removeTab(size-1-index)

        titles = self.getTitles()

        self.tab = []
        self.plot_canvas = []

        for index in range(0, len(titles)):
            self.tab.append(oasysgui.createTabPage(self.tabs, titles[index]))
            self.plot_canvas.append(None)

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

    def getDefaultPlotTabIndex(self):
        return -1

    def getTitles(self):
        return ["Calculation Result"]

    def getXTitles(self):
        return ["Energy [eV]"]

    def getYTitles(self):
        return ["X [$\mu$m]"]

    def getVariablesToPlot(self):
        return [(0, 1)]

    def getLogPlot(self):
        return [(False, False)]

    def set_ViewType(self):
        self.progressBarInit()

        if not self.calculated_data==None:
            try:
                self.initializeTabs()

                self.plot_results(self.calculated_data)
            except Exception as exception:
                QtWidgets.QMessageBox.critical(self, "Error",
                                            str(exception),
                                            QtWidgets.QMessageBox.Ok)

        self.progressBarFinished()

    def plot_results(self, calculated_data, progressBarValue=80):
        if not self.view_type == 0:
            if not calculated_data is None:
                current_index = self.tabs.currentIndex()

                self.view_type_combo.setEnabled(False)

                esrf_data = calculated_data.get_content("esrf_data")

                titles = self.getTitles()
                xtitles = self.getXTitles()
                ytitles = self.getYTitles()

                progress_bar_step = (100-progressBarValue)/len(titles)

                for index in range(0, len(titles)):
                    x_index, y_index = self.getVariablesToPlot()[index]
                    log_x, log_y = self.getLogPlot()[index]

                    try:
                        self.plot_histo(esrf_data[:, x_index],
                                        esrf_data[:, y_index],
                                        progressBarValue + ((index+1)*progress_bar_step),
                                        tabs_canvas_index=index,
                                        plot_canvas_index=index,
                                        title=titles[index],
                                        xtitle=xtitles[index],
                                        ytitle=ytitles[index],
                                        log_x=log_x,
                                        log_y=log_y)

                        # self.tabs.setCurrentIndex(index)
                    except Exception as e:
                        self.view_type_combo.setEnabled(True)

                        raise Exception("Data not plottable: bad content\n" + str(e))

                self.view_type_combo.setEnabled(True)

                try:
                    self.tabs.setCurrentIndex(current_index)
                except:
                    if self.getDefaultPlotTabIndex() == -1:
                        self.tabs.setCurrentIndex(len(titles) - 1)
                    else:
                        self.tabs.setCurrentIndex(self.getDefaultPlotTabIndex())


            else:
                raise Exception("Empty Data")

    def writeStdOut(self, text):
        cursor = self.esrf_output.textCursor()
        cursor.movePosition(QtGui.QTextCursor.End)
        cursor.insertText(text)
        self.esrf_output.setTextCursor(cursor)
        self.esrf_output.ensureCursorVisible()

    def plot_histo(self, x, y, progressBarValue, tabs_canvas_index, plot_canvas_index, title="", xtitle="", ytitle="",
                   log_x=False, log_y=False, color='blue', replace=True, control=False):


        if self.plot_canvas[plot_canvas_index] is None:
            self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(parent=None,
                                                                      backend=None,
                                                                      resetzoom=True,
                                                                      autoScale=False,
                                                                      logScale=True,
                                                                      grid=True,
                                                                      curveStyle=True,
                                                                      colormap=False,
                                                                      aspectRatio=False,
                                                                      yInverted=False,
                                                                      copy=True,
                                                                      save=True,
                                                                      print_=True,
                                                                      control=control,
                                                                      position=True,
                                                                      roi=False,
                                                                      mask=False,
                                                                      fit=False)
            self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
            self.plot_canvas[plot_canvas_index].setActiveCurveColor(color="#00008B")
            self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
            self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)

            self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])
        ESRFWidget.plot_histo_2(self.plot_canvas[plot_canvas_index], x, y, title, xtitle, ytitle, color, replace)

        self.plot_canvas[plot_canvas_index].setXAxisLogarithmic(log_x)
        self.plot_canvas[plot_canvas_index].setYAxisLogarithmic(log_y)

        if min(y) < 0:
            if log_y:
                self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y)*1.2, max(y)*1.2)
            else:
                self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y)*1.01, max(y)*1.01)
        else:
            if log_y:
                self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y), max(y)*1.2)
            else:
                self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y), max(y)*1.01)


        self.progressBarSet(progressBarValue)

    def plot_data1D(self, dataX, dataY, tabs_canvas_index, plot_canvas_index, title="", xtitle="", ytitle=""):

        self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(0))

        self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow()

        self.plot_canvas[plot_canvas_index].addCurve(dataX, dataY,)

        self.plot_canvas[plot_canvas_index].resetZoom()
        self.plot_canvas[plot_canvas_index].setXAxisAutoScale(True)
        self.plot_canvas[plot_canvas_index].setYAxisAutoScale(True)
        self.plot_canvas[plot_canvas_index].setGraphGrid(False)

        self.plot_canvas[plot_canvas_index].setXAxisLogarithmic(False)
        self.plot_canvas[plot_canvas_index].setYAxisLogarithmic(False)
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.plot_canvas[plot_canvas_index].setGraphTitle(title)

        self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])

    def plot_data2D(self, data2D, dataX, dataY, tabs_canvas_index, plot_canvas_index, title="", xtitle="", ytitle="", mode=2):

        for i in range(1+self.tab[tabs_canvas_index].layout().count()):
            self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(i))

        if mode == 0:
            figure = FigureCanvas(gol.plot_image(data2D,
                                                 dataX,
                                                 dataY,
                                                 xtitle=xtitle,
                                                 ytitle=ytitle,
                                                 title=title,
                                                 show=False,
                                                 aspect='auto'))


            self.plot_canvas[plot_canvas_index] = figure
        else:

            origin = (dataX[0],dataY[0])
            scale = (dataX[1]-dataX[0],dataY[1]-dataY[0])

            data_to_plot = data2D.T

            colormap = {"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256}


            if mode == 1:
                #TODO: delete: srio commented this part as it is never used
                raise Exception("Cannot use ImageView")
            elif mode == 2:

                self.plot_canvas[plot_canvas_index] = Plot2D()

                self.plot_canvas[plot_canvas_index].resetZoom()
                self.plot_canvas[plot_canvas_index].setXAxisAutoScale(True)
                self.plot_canvas[plot_canvas_index].setYAxisAutoScale(True)
                self.plot_canvas[plot_canvas_index].setGraphGrid(False)
                self.plot_canvas[plot_canvas_index].setKeepDataAspectRatio(True)
                self.plot_canvas[plot_canvas_index].yAxisInvertedAction.setVisible(False)

                self.plot_canvas[plot_canvas_index].setXAxisLogarithmic(False)
                self.plot_canvas[plot_canvas_index].setYAxisLogarithmic(False)
                #silx 0.4.0
                self.plot_canvas[plot_canvas_index].getMaskAction().setVisible(False)
                self.plot_canvas[plot_canvas_index].getRoiAction().setVisible(False)
                self.plot_canvas[plot_canvas_index].getColormapAction().setVisible(False)
                self.plot_canvas[plot_canvas_index].setKeepDataAspectRatio(False)



                self.plot_canvas[plot_canvas_index].addImage(numpy.array(data_to_plot),
                                                             legend="zio billy",
                                                             scale=scale,
                                                             origin=origin,
                                                             colormap=colormap,
                                                             replace=True)

                self.plot_canvas[plot_canvas_index].setActiveImage("zio billy")

            self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
            self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
            self.plot_canvas[plot_canvas_index].setGraphTitle(title)

        self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])



    def plot_data3D(self, data3D, dataE, dataX, dataY, tabs_canvas_index, plot_canvas_index, title="", xtitle="", ytitle=""):

        for i in range(1+self.tab[tabs_canvas_index].layout().count()):
            self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(i))

        #self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(0))


        xmin = numpy.min(dataX)
        xmax = numpy.max(dataX)
        ymin = numpy.min(dataY)
        ymax = numpy.max(dataY)


        stepX = dataX[1]-dataX[0]
        stepY = dataY[1]-dataY[0]
        if len(dataE) > 1: stepE = dataE[1]-dataE[0]
        else: stepE = 1.0

        if stepE == 0.0: stepE = 1.0
        if stepX == 0.0: stepX = 1.0
        if stepY == 0.0: stepY = 1.0

        dim0_calib = (dataE[0],stepE)
        dim1_calib = (ymin, stepY)
        dim2_calib = (xmin, stepX)


        data_to_plot = numpy.swapaxes(data3D,1,2)

        colormap = {"name":"temperature", "normalization":"linear", "autoscale":True, "vmin":0, "vmax":0, "colors":256}

        self.plot_canvas[plot_canvas_index] = StackViewMainWindow()

        self.plot_canvas[plot_canvas_index].setGraphTitle(title)
        self.plot_canvas[plot_canvas_index].setLabels(["Photon Energy [eV]",ytitle,xtitle])
        self.plot_canvas[plot_canvas_index].setColormap(colormap=colormap)

        self.plot_canvas[plot_canvas_index].setStack(numpy.array(data_to_plot),
                                                     calibrations=[dim0_calib, dim1_calib, dim2_calib] )
        self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])



    @classmethod
    def plot_histo_2(cls, plot_window, x, y, title, xtitle, ytitle, color='blue', replace=True):
        matplotlib.rcParams['axes.formatter.useoffset']='False'
        plot_window.addCurve(x, y, title, symbol='', color=color, xlabel=xtitle, ylabel=ytitle, replace=replace) #'+', '^', ','
        if not xtitle is None: plot_window.setGraphXLabel(xtitle)
        if not ytitle is None: plot_window.setGraphYLabel(ytitle)
        if not title is None: plot_window.setGraphTitle(title)

        # plot_window.setDrawModeEnabled(True, 'rectangle')
        plot_window.setInteractiveMode('zoom',color='orange')
        plot_window.resetZoom()
        plot_window.replot()

        plot_window.setActiveCurve(title)


    def compute(self):
        self.setStatusMessage("Running Command")

        self.progressBarInit()

        try:
            self.esrf_output.setText("")

            sys.stdout = EmittingStream(textWritten=self.writeStdOut)

            self.progressBarSet(20)

            self.check_fields()

            calculation_output = self.do_calculation()

            self.progressBarSet(50)

            if calculation_output is None:
                raise Exception("No result")
            else:
                self.calculated_data = self.extract_data_from_calculation_output(calculation_output)

                self.add_specific_content_to_calculated_data(self.calculated_data)

                self.setStatusMessage("Plotting Results")

                self.plot_results(self.calculated_data, progressBarValue=60)

                self.setStatusMessage("")

                self.send("esrf_data", self.calculated_data)

        except Exception as exception:
            QtWidgets.QMessageBox.critical(self, "Error",
                                       str(exception), QtWidgets.QMessageBox.Ok)

            self.setStatusMessage("Error!")

            if self.IS_DEVELOP: raise exception

        self.progressBarFinished()


    def defaults(self):
         self.resetSettings()

    def help1(self):
        raise Exception("This method should be reimplementd in subclasses!")

    def get_help_name(self):
        raise Exception("This method should be reimplementd in subclasses!")

    def check_fields(self):
        raise Exception("This method should be reimplementd in subclasses!")

    def do_calculation(self):
        raise Exception("This method should be reimplementd in subclasses!")

    def extract_data_from_calculation_output(self, calculation_output):
        calculated_data = DataExchangeObject("ESRF", self.get_data_exchange_widget_name())

        calculated_data.add_content("esrf_data", calculation_output)

        return calculated_data

    def get_data_exchange_widget_name(self):
        raise Exception("This method should be reimplementd in subclasses!")


    def add_specific_content_to_calculated_data(self, calculated_data):
        pass





if __name__ == "__main__":
    a = QApplication(sys.argv)
    ow = ESRFWidget()
    ow.show()
    a.exec_()
    ow.saveSettings()



from orangewidget.settings import Setting
from orangewidget import gui

from oasys.widgets import gui as oasysgui

from orangecontrib.esrf.wofry.widgets.gui.ow_optical_element_1d import OWWOOpticalElement1D
from orangecontrib.esrf.wofry.util.toolbox import WOToolbox1D #TODO from wofryimpl....

class OWWOToolbox1D(OWWOOpticalElement1D):

    name = "Toolbox Wavefront 1D"
    description = "Wofry: Toolbox Wavefront 1D"
    icon = "icons/util.png"
    priority = 150

    crop_factor = Setting(1.0)
    shift_center_in_microns = Setting(0.0)
    abscissas_factor = Setting(1.0)
    change_photon_energy = Setting(0)    # 0=No, 1=Yes
    new_photon_energy    = Setting(0.0)  # if change_photon_energy=1, the new photon energy in eV

    def __init__(self):

        super().__init__(is_automatic=True, show_view_options=True, show_script_tab=True)

    def draw_specific_box(self):

        toolbox_box = oasysgui.widgetBox(self.tab_bas, "Toolbox 1D Setting", addSpace=False, orientation="vertical",
                                           height=350)

        tmp = oasysgui.lineEdit(toolbox_box, self, "shift_center_in_microns", "Shift center [microns]",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("shift_center_in_microns")


        tmp = oasysgui.lineEdit(toolbox_box, self, "abscissas_factor", "Multiply abscissas by",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("abscissas_factor")

        tmp = oasysgui.lineEdit(toolbox_box, self, "crop_factor", "Crop/pad factor (<1:crop, >1:pad) ",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("crop_factor")


        gui.comboBox(toolbox_box, self, "change_photon_energy", label="Change photon energy",
                     items=['No','Yes'], #callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")


        new_energy_box = oasysgui.widgetBox(toolbox_box, "", addSpace=False,
                                        orientation="horizontal")  # width=550, height=50)
        tmp = oasysgui.lineEdit(new_energy_box, self, "new_photon_energy", "new photon energy",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("new_photon_energy")
        self.show_at("self.change_photon_energy == 1", new_energy_box)



    def get_optical_element(self):

        return WOToolbox1D(name=self.oe_name,
                           crop_factor=self.crop_factor,
                           shift_center=self.shift_center_in_microns*1e-6,
                           abscissas_factor=self.abscissas_factor,
                           change_photon_energy=self.change_photon_energy,
                           new_photon_energy=self.new_photon_energy)

    def check_data(self):
        super().check_data()
        # congruence.checkFileName(self.file_with_thickness_mesh)

    def receive_specific_syned_data(self, optical_element):
        pass

if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    def get_example_wofry_data():
        from wofryimpl.propagator.light_source import WOLightSource
        from wofryimpl.beamline.beamline import WOBeamline
        from orangecontrib.wofry.util.wofry_objects import WofryData
        light_source = WOLightSource(dimension=1,
                                     initialize_from=2,
                                     sigma_h=.1e-6,
                                     sigma_v=.1e-6,
                                     )

        return WofryData(wavefront=light_source.get_wavefront(),
                           beamline=WOBeamline(light_source=light_source))



    a = QApplication(sys.argv)
    ow = OWWOToolbox1D()
    ow.set_input(get_example_wofry_data())

    ow.show()
    a.exec_()
    ow.saveSettings()



import numpy
import os, sys

import orangecanvas.resources as resources

from PyQt5.QtGui import QPalette, QColor, QFont, QPixmap
from PyQt5.QtWidgets import QMessageBox, QLabel, QSizePolicy
from PyQt5.QtCore import Qt

from orangewidget import gui
from orangewidget.settings import Setting

from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import TriggerIn, TriggerOut, EmittingStream

from orangecontrib.esrf.syned.util.lens import Lens # TODO: from syned.beamline.optical_elements....

from syned.widget.widget_decorator import WidgetDecorator

from orangecontrib.wofry.util.wofry_objects import WofryData

from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D

from orangecontrib.esrf.wofry.widgets.gui.ow_optical_element_1d import OWWOOpticalElement1D # TODO rom orangecontrib.wofry.widgets.gui.ow_optical_element_1d

from orangecontrib.esrf.wofry.util.lens import WOLens1D


class OWWORealLens1D(OWWOOpticalElement1D):

    name = "Real Lens 1D"
    id = "WofryLens1D"
    description = "Wofry: Real Lens 1D"
    icon = "icons/lens1d.png"
    priority = 4

    category = "Wofry Wavefront Propagation"
    keywords = ["data", "file", "load", "read"]

    outputs = [{"name":"WofryData",
                "type":WofryData,
                "doc":"WofryData",
                "id":"WofryData"},
               {"name":"Trigger",
                "type": TriggerIn,
                "doc":"Feedback signal to start a new beam simulation",
                "id":"Trigger"}]

    inputs = [("WofryData", WofryData, "set_input"),
              ("DABAM 1D Profile", numpy.ndarray, "receive_dabam_profile"),
              ("Trigger", TriggerOut, "receive_trigger_signal"),
              WidgetDecorator.syned_input_data()[0]]


    # Basic lens parameters____________________________________________________________________________________________#
    shape = Setting(1)
    radius = Setting(0.0005)
    wall_thickness = Setting(0.00005)
    lens_aperture = Setting(0.001)
    # number_of_refractive_surfaces = Setting(1)
    number_of_curved_surfaces = Setting(2)
    n_lenses = Setting(1)
    material = Setting(0)
    refraction_index_delta = Setting(5.3e-7)
    att_coefficient = Setting(0.00357382)
    error_flag = Setting(0)
    error_file = Setting("<none>")
    error_edge_management = Setting(0)
    write_profile_flag = Setting(0)
    write_profile = Setting("profile1D.dat")

    image1_path = os.path.join(resources.package_dirname("orangecontrib.esrf.wofry.widgets.gui"), "misc", "Refractor_parameters.png")

    # Lens misaligments for Barc4ro____________________________________________________________________________________#

    mis_flag = Setting(0)
    xc = Setting(0.0)
    ang_rot = Setting(0.0)
    wt_offset_ffs = Setting(0.0)
    offset_ffs = Setting(0.0)
    tilt_ffs = Setting(0.0)
    wt_offset_bfs = Setting(0.0)
    offset_bfs = Setting(0.0)
    tilt_bfs = Setting(0.0)

    image2_path = os.path.join(resources.package_dirname("orangecontrib.esrf.wofry.widgets.gui"), "misc", "Refractor_misalignments.png")

    input_data = None

    def __init__(self):
        super().__init__(is_automatic=True, show_view_options=True, show_script_tab=True)

    def draw_specific_box(self):

        box_refractor = oasysgui.widgetBox(self.tab_bas, "1D Lens Geometry", addSpace=False, orientation="vertical",
                                           height=350)


        gui.comboBox(box_refractor, self, "shape", label="Lens shape", labelWidth=350,
                     items=["Flat", "Parabolic", "Circular"],
                     sendSelectedValue=False, orientation="horizontal",callback=self.set_visible)

        self.box_radius_id = oasysgui.widgetBox(box_refractor, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_radius_id, self, "radius", "(R) Radius of curvature [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("radius")

        self.box_aperture_id = oasysgui.widgetBox(box_refractor, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_aperture_id, self, "lens_aperture", "(A) Lens aperture [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("lens_aperture")

        self.box_wall_thickness_id = oasysgui.widgetBox(box_refractor, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_wall_thickness_id, self, "wall_thickness", "(t_wall) Wall thickness [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("wall_thickness")

        # Help figure
        self.figure_box1 = oasysgui.widgetBox(box_refractor, "Principal parabolic lens parameters", addSpace=False, orientation="horizontal")
        label1 = QLabel("")
        label1.setAlignment(Qt.AlignCenter)
        label1.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label1.setPixmap(QPixmap(self.image1_path))
        self.figure_box1.layout().addWidget(label1)

        #
        # Tab Lens pars
        #
        self.tab_err = oasysgui.createTabPage(self.tabs_setting, "Parms")

        # Box refractor:_________________________________________________________________________

        box_refractor2 = oasysgui.widgetBox(self.tab_err, "1D Lens parameters", addSpace=False, orientation="vertical")


        # gui.comboBox(box_refractor2, self, "number_of_refractive_surfaces", label="Number of refractive surfaces", labelWidth=350,
        #              items=["1", "2"],
        #              sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(box_refractor2, self, "number_of_curved_surfaces", label="Number of curved surfaces", labelWidth=350,
                     items=["0 (parallel plate)", "1 (plano-concave)", "2 (bi-concave)"],
                     sendSelectedValue=False, orientation="horizontal", callback=self.set_visible)


        self.box_n_lenses_id = oasysgui.widgetBox(box_refractor2, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_n_lenses_id, self, "n_lenses", "Number of lenses",
                          labelWidth=300, valueType=int, orientation="horizontal")
        tmp.setToolTip("n_lenses")

        gui.comboBox(box_refractor2, self, "material", label="Lens material",
                     items=self.get_material_name(),
                     callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        self.box_refraction_index_id = oasysgui.widgetBox(box_refractor2, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_refraction_index_id, self, "refraction_index_delta", "Refraction index delta",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("refraction_index_delta")

        self.box_att_coefficient_id = oasysgui.widgetBox(box_refractor2, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_att_coefficient_id, self, "att_coefficient", "Attenuation coefficient [m-1]",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("att_coefficient")


        gui.comboBox(box_refractor2, self, "write_profile_flag", label="Dump profile to file",
                     items=["No", "Yes"], sendSelectedValue=False, orientation="horizontal",
                     callback=self.set_visible)

        self.box_file_out = gui.widgetBox(box_refractor2, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.box_file_out, self, "write_profile", "File name",
                            labelWidth=200, valueType=str, orientation="horizontal")
        # error box _________________________________________________________________________

        box_errors = oasysgui.widgetBox(self.tab_err, "1D Lens error profile", addSpace=False, orientation="vertical")


        gui.comboBox(box_errors, self, "error_flag", label="Add profile deformation",
                     items=["No", "Yes (from file)"],
                     callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        self.error_profile = oasysgui.widgetBox(box_errors, "", addSpace=True, orientation="vertical")
        file_box_id = oasysgui.widgetBox(self.error_profile, "", addSpace=True, orientation="horizontal")

        self.error_file_id = oasysgui.lineEdit(file_box_id, self, "error_file", "Error file X[m] Y[m]",
                                                    labelWidth=100, valueType=str, orientation="horizontal")
        gui.button(file_box_id, self, "...", callback=self.set_error_file)
        self.error_file_id.setToolTip("error_file")

        gui.comboBox(self.error_profile, self, "error_edge_management", label="Manage edges",
                     items=["Extrapolate deformation profile", "Crop beam to deformation profile dimension"],
                     callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        #
        #
        #
        self.tab_mis = oasysgui.createTabPage(self.tabs_setting, "Misalignments")

        box_misaligments = oasysgui.widgetBox(self.tab_mis, "Typical lens misalignments (only for parabolic shape)",
                                              addSpace=False, orientation="vertical")

        # Tab Misalignments: Typical lens misalignments for Barc4ro_temporarily_only_for_parabolic______________________

        self.mis_flag_box = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="vertical")

        gui.comboBox(self.mis_flag_box, self, "mis_flag", label="Add misalignments", labelWidth=350,
                     items=["No","Yes"], sendSelectedValue=False, orientation="horizontal", callback=self.set_visible)

        self.box_xc_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_xc_id, self, "xc", "(xc) Coordinate of center [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("lens_aperture")

        self.box_ang_rot_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_ang_rot_id, self, "ang_rot", "(ang_rot) Angle of full CRL rotation about horizontal [rad]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("ang_rot")

        self.box_wt_offset_ffs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_wt_offset_ffs_id, self, "wt_offset_ffs", "(wt_offset_ffs) ffs offset respect wall thickness [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("wt_offset_ffs")

        self.box_offset_ffs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_offset_ffs_id, self, "offset_ffs", "(offset_ffs) Lateral ffs offset respect horizontal [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("offset_ffs")

        self.box_tilt_ffs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_tilt_ffs_id, self, "tilt_ffs", "(tilt_ffs) Angle of the parabolic ffs respect horizontal [rad]",
                          labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("tilt_ffs")

        self.box_wt_offset_bfs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_wt_offset_bfs_id, self, "wt_offset_bfs",
                                "(wt_offset_bfs) bfs offset respect wall thickness [m]",
                                labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("wt_offset_bfs")

        self.box_offset_bfs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_offset_bfs_id, self, "offset_bfs",
                                "(offset_bfs) Lateral bfs offset respect horizontal [m]",
                                labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("offset_bfs")

        self.box_tilt_bfs_id = oasysgui.widgetBox(box_misaligments, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_tilt_bfs_id, self, "tilt_bfs",
                                "(tilt_bfs) Angle of the parabolic bfs respect horizontal [rad]",
                                labelWidth=300, valueType=float, orientation="horizontal")
        tmp.setToolTip("tilt_bfs")

        self.figure_box2 = oasysgui.widgetBox(box_misaligments, "Front focusing surface (ffs), back focusing surface (bfs)", addSpace=False, orientation="horizontal")

        label2 = QLabel("")
        label2.setAlignment(Qt.AlignCenter)
        label2.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label2.setPixmap(QPixmap(self.image2_path))
        self.figure_box2.layout().addWidget(label2)

        self.set_visible()

    def get_material_name(self, index=None):
        materials_list = ["External", "Be", "Al", "Diamond"]
        if index is None:
            return materials_list
        else:
            return materials_list[index]

    def set_visible(self):
        self.error_profile.setVisible(self.error_flag)
        self.box_radius_id.setVisible(self.shape in [1,2])
        self.box_refraction_index_id.setVisible(self.material in [0])
        self.box_att_coefficient_id.setVisible(self.material in [0])
        self.box_file_out.setVisible(self.write_profile_flag == 1)

        # Variables misalignments #
        self.mis_flag_box.setVisible(self.shape in [1]) # Only for parabolic shape
        self.box_xc_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_ang_rot_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_wt_offset_ffs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_offset_ffs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_tilt_ffs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_wt_offset_bfs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_offset_bfs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.box_tilt_bfs_id.setVisible(self.mis_flag in [1] and self.shape in [1])
        self.figure_box2.setVisible(self.mis_flag in [1] and self.shape in [1])

    def set_error_file(self):
        self.error_file_id.setText(oasysgui.selectFileFromDialog(self, self.error_file, "Open file with profile error"))

    def check_fields(self):

        self.radius = congruence.checkStrictlyPositiveNumber(self.radius, "Radius")
        self.wall_thickness = congruence.checkNumber(self.wall_thickness, "Wall thickness")
        self.lens_aperture = congruence.checkNumber(self.lens_aperture, "Lens aperture")
        self.n_lenses = congruence.checkStrictlyPositiveNumber(self.n_lenses, "Number of Lenses")
        self.refraction_index_delta = congruence.checkNumber(self.refraction_index_delta, "Refraction index delta")
        self.att_coefficient = congruence.checkNumber(self.att_coefficient, "Attenuation coefficient")
        self.error_file = congruence.checkFileName(self.error_file)
        self.n_lenses = congruence.checkNumber(self.n_lenses, "Number of lenses")

        self.xc = congruence.checkNumber(self.xc, "xc")
        self.ang_rot = congruence.checkNumber(self.ang_rot, "ang_rot")
        self.wt_offset_ffs = congruence.checkNumber(self.wt_offset_ffs, "wt_offset_ffs")
        self.offset_ffs = congruence.checkNumber(self.offset_ffs, "offset_ffs")
        self.tilt_ffs = congruence.checkNumber(self.tilt_ffs, "tilt_ffs")
        self.wt_offset_bfs = congruence.checkNumber(self.offset_bfs, "offset_ffs")
        self.offset_bfs = congruence.checkNumber(self.offset_bfs, "offset_bfs")
        self.tilt_bfs = congruence.checkNumber(self.tilt_bfs, "tilt_bfs")

    def receive_trigger_signal(self, trigger):

        if trigger and trigger.new_object == True:
            if trigger.has_additional_parameter("variable_name"):
                variable_name = trigger.get_additional_parameter("variable_name").strip()
                variable_display_name = trigger.get_additional_parameter("variable_display_name").strip()
                variable_value = trigger.get_additional_parameter("variable_value")
                variable_um = trigger.get_additional_parameter("variable_um")

                if "," in variable_name:
                    variable_names = variable_name.split(",")

                    for variable_name in variable_names:
                        setattr(self, variable_name.strip(), variable_value)
                else:
                    setattr(self, variable_name, variable_value)

                self.propagate_wavefront()

    def receive_syned_data(self):
        raise Exception(NotImplementedError)



    def receive_dabam_profile(self, dabam_profile):
        if not dabam_profile is None:
            try:
                file_name = "dabam_profile_" + str(id(self)) + ".dat"

                file = open(file_name, "w")

                for element in dabam_profile:
                    file.write(str(element[0]) + " " + str(element[1]) + "\n")

                file.flush()
                file.close()

                self.error_flag = 1
                self.error_file = file_name
                self.set_visible()

            except Exception as exception:
                QMessageBox.critical(self, "Error", exception.args[0], QMessageBox.Ok)

                if self.IS_DEVELOP: raise exception

    def get_optical_element(self):
        return WOLens1D.create_from_keywords(
                    name                         =self.oe_name,
                    shape                        = self.shape,
                    radius                       = self.radius,
                    lens_aperture                = self.lens_aperture,
                    wall_thickness               = self.wall_thickness,
                    material                     = self.get_material_name(self.material),
                    refraction_index_delta       = self.refraction_index_delta,
                    att_coefficient              = self.att_coefficient,
                    number_of_curved_surfaces    = self.number_of_curved_surfaces,
                    n_lenses                     = self.n_lenses,
                    error_flag                   = self.error_flag,
                    error_file                   = self.error_file,
                    error_edge_management        = self.error_edge_management,
                    write_profile_flag           = self.write_profile_flag,
                    write_profile                = self.write_profile,
                    mis_flag                     = self.mis_flag,
                    xc                           = self.xc,
                    ang_rot                      = self.ang_rot,
                    wt_offset_ffs                = self.wt_offset_ffs,
                    offset_ffs                   = self.offset_ffs,
                    tilt_ffs                     = self.tilt_ffs,
                    wt_offset_bfs                = self.wt_offset_bfs,
                    offset_bfs                   = self.offset_bfs,
                    tilt_bfs                     = self.tilt_bfs,
                    )

    #
    # overwritten methods to append profile plot
    #

    def get_titles(self):
        titles = super().get_titles()
        titles.append("O.E. Profile")
        return titles

    def do_plot_results(self, progressBarValue=80): # OVERWRITTEN

        super().do_plot_results(progressBarValue, closeProgressBar=False)
        if not self.view_type == 0:
            if not self.wavefront_to_plot is None:

                self.progressBarSet(progressBarValue)

                wo_lens = self.get_optical_element()
                abscissas_on_lens, lens_thickness = wo_lens.get_surface_thickness_mesh(self.input_data.get_wavefront())

                self.plot_data1D(x=abscissas_on_lens*1e6, #TODO check how is possible to plot both refractive surfaces
                                 y=lens_thickness*1e6, # in microns
                                 progressBarValue=progressBarValue + 10,
                                 tabs_canvas_index=4,
                                 plot_canvas_index=4,
                                 calculate_fwhm=False,
                                 title=self.get_titles()[4],
                                 xtitle="Spatial Coordinate along o.e. [$\mu$m]",
                                 ytitle="Total lens thickness [$\mu$m]")

                self.progressBarFinished()


if __name__ == '__main__':

    from PyQt5.QtWidgets import QApplication

    def create_wavefront():
        #
        # create input_wavefront
        #
        from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
        input_wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.0005, x_max=0.0005,
                                                                             number_of_points=1000)
        input_wavefront.set_photon_energy(10000)
        input_wavefront.set_spherical_wave(radius=13.73, center=0, complex_amplitude=complex(1, 0))
        return input_wavefront

    def get_example_wofry_data():
        from wofryimpl.propagator.light_source import WOLightSource
        from wofryimpl.beamline.beamline import WOBeamline
        from orangecontrib.wofry.util.wofry_objects import WofryData

        light_source = WOLightSource(dimension=1,
                                     kind_of_wave=1,
                                     initialize_from=0,
                                     range_from_h=-0.0005,
                                     range_to_h=0.0005,
                                     range_from_v=-0.0005,
                                     range_to_v=0.0005,
                                     number_of_points_h=1000,
                                     number_of_points_v=1,
                                     energy=10000.0,
                                     radius=13.73,
                                     )
        return WofryData(wavefront=light_source.get_wavefront(),
                           beamline=WOBeamline(light_source=light_source))

    app = QApplication([])
    ow = OWWORealLens1D()
    # ow.set_input(create_wavefront())
    ow.set_input(get_example_wofry_data())
    # ow.receive_dabam_profile(numpy.array([[-1.50,0],[1.50,0]]))

    ow.propagate_wavefront()

    ow.show()
    app.exec_()
    ow.saveSettings()


import numpy
from syned.storage_ring.light_source import LightSource
from syned.storage_ring.electron_beam import ElectronBeam
from syned.storage_ring.magnetic_structures.undulator import Undulator

from wofry.beamline.decorators import LightSourceDecorator
from wofryimpl.propagator.util.undulator_coherent_mode_decomposition_1d import UndulatorCoherentModeDecomposition1D

from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D

class WOLightSourceCMD(LightSource, LightSourceDecorator, UndulatorCoherentModeDecomposition1D):
    def __init__(self,
                 name                = "Undefined",
                 # electron_beam       = None,
                 # magnetic_structure  = None,
                 undulator_coherent_mode_decomposition_1d = None,
                 dimension           = 1,
                 ):

        electron_beam = ElectronBeam(energy_in_GeV=undulator_coherent_mode_decomposition_1d.electron_energy,
                                     current=undulator_coherent_mode_decomposition_1d.electron_current)
        magnetic_structure = Undulator(K_vertical=undulator_coherent_mode_decomposition_1d.K,
                                       period_length=undulator_coherent_mode_decomposition_1d.undulator_period,
                                       number_of_periods=undulator_coherent_mode_decomposition_1d.undulator_nperiods)

        LightSource.__init__(self, name=name, electron_beam=electron_beam, magnetic_structure=magnetic_structure)
        UndulatorCoherentModeDecomposition1D.__init__(self,
                 electron_energy      = undulator_coherent_mode_decomposition_1d.electron_energy     ,
                 electron_current     = undulator_coherent_mode_decomposition_1d.electron_current    ,
                 undulator_period     = undulator_coherent_mode_decomposition_1d.undulator_period    ,
                 undulator_nperiods   = undulator_coherent_mode_decomposition_1d.undulator_nperiods  ,
                 K                    = undulator_coherent_mode_decomposition_1d.K                   ,
                 photon_energy        = undulator_coherent_mode_decomposition_1d.photon_energy       ,
                 abscissas_interval   = undulator_coherent_mode_decomposition_1d.abscissas_interval  ,
                 number_of_points     = undulator_coherent_mode_decomposition_1d.number_of_points    ,
                 distance_to_screen   = undulator_coherent_mode_decomposition_1d.distance_to_screen  ,
                 scan_direction       = undulator_coherent_mode_decomposition_1d.scan_direction      ,
                 magnification_x      = undulator_coherent_mode_decomposition_1d.magnification_x     ,
                 sigmaxx              = 1.0 / numpy.sqrt(undulator_coherent_mode_decomposition_1d.mxx)  ,
                 sigmaxpxp            = 1.0 / numpy.sqrt(undulator_coherent_mode_decomposition_1d.mxpxp),
                 useGSMapproximation  = undulator_coherent_mode_decomposition_1d.useGSMapproximation ,
                )

        self._dimension =  dimension
        self.dimension = dimension
        self._set_support_text([
                    # ("name"      ,           "to define ", "" ),
                    ("dimension"      , "dimension ", "" ),
            ] )

    def get_dimension(self):
        return self._dimension

    # from Wofry Decorator
    def get_wavefront(self):
        self.get_eigenvector_wavefront(mode=0)

    def to_python_code(self, do_plot=True, add_import_section=False):

        txt = ""

        txt += "#"
        txt += "\n# create output_wavefront\n#"
        txt += "\n#"

        if self._dimension == 1:
            txt += "\nfrom wofryimpl.propagator.util.undulator_coherent_mode_decomposition_1d import UndulatorCoherentModeDecomposition1D"

        else:
            raise Exception("Not implemented")

        txt += "\ncoherent_mode_decomposition = UndulatorCoherentModeDecomposition1D("
        txt += "\n    electron_energy=%g," % self.electron_energy
        txt += "\n    electron_current=%g," % self.electron_current
        txt += "\n    undulator_period=%g," % self.undulator_period
        txt += "\n    undulator_nperiods=%g," % self.undulator_nperiods
        txt += "\n    K=%g," % self.K
        txt += "\n    photon_energy=%g," % self.photon_energy
        txt += "\n    abscissas_interval=%g," % self.abscissas_interval
        txt += "\n    number_of_points=%g," % self.number_of_points
        txt += "\n    distance_to_screen=%g," % self.distance_to_screen
        txt += "\n    scan_direction='%s'," % self.scan_direction
        txt += "\n    sigmaxx=%g,"   % (1.0 / numpy.sqrt(self.mxx))
        txt += "\n    sigmaxpxp=%g," % (1.0 / numpy.sqrt(self.mxpxp))
        if self.useGSMapproximation:
            txt += "\n    useGSMapproximation=True,)"
        else:
            txt += "\n    useGSMapproximation=False,)"
        txt += "\n# make calculation"
        txt += "\ncoherent_mode_decomposition_results = coherent_mode_decomposition.calculate()"

        txt += "\n\nmode_index = 0"
        txt += "\noutput_wavefront = coherent_mode_decomposition.get_eigenvector_wavefront(mode_index)"


        return txt


class WOPySRULightSource(LightSource, LightSourceDecorator):
    def __init__(self,
                 name                = "Undefined",
                 electron_beam       = None,
                 magnetic_structure  = None,
                 ):


        LightSource.__init__(self, name=name, electron_beam=electron_beam, magnetic_structure=magnetic_structure)

        self.__source_wavefront_parameters = {
            'distance'       : 30.0,
            'gapH'           : 0.003,
            'gapV'           : 0.003,
            'photon_energy'  : 7000.0,
            'h_slit_points'    : 51,
            'v_slit_points'    : 51,
            'flag_send_wavefront_dimension' : 0, # 0=2D wavefront, 1=H wawefront, 2=V wavefront (for python script)
            }


        self._dimension =  2

        self._set_support_text([
                    # ("name"      ,           "to define ", "" ),
                    ("dimension"      , "dimension ", "" ),
            ] )



    @classmethod
    def initialize_from_keywords(cls,
                                 name="Undefined",
                                 energy_in_GeV=6.04,
                                 current=0.2,
                                 K_vertical=1.68,
                                 period_length=0.018,
                                 number_of_periods=222,
                                 distance=10.0,
                                 gapH=0.003,
                                 gapV=0.003,
                                 photon_energy=7000.0,
                                 h_slit_points=51,
                                 v_slit_points=51,
                                 flag_send_wavefront_dimension=0,
                                 ):


        out = WOPySRULightSource(name=name,
                                 electron_beam=ElectronBeam(
                                     energy_in_GeV=energy_in_GeV,
                                     current=current),
                                 magnetic_structure=Undulator(
                                     K_vertical=K_vertical,
                                     period_length=period_length,
                                     number_of_periods=number_of_periods))

        out.set_source_wavefront_parameters(distance=distance,
                                            gapH=gapH,
                                            gapV=gapV,
                                            photon_energy=photon_energy,
                                            h_slit_points=h_slit_points,
                                            v_slit_points=v_slit_points,
                                            flag_send_wavefront_dimension=flag_send_wavefront_dimension)

        return out


    def set_source_wavefront_parameters(self, distance=None, gapH=None, gapV=None, photon_energy=None,
                                        h_slit_points=None, v_slit_points=None,
                                        flag_send_wavefront_dimension=None,):

        if distance is not None: self.__source_wavefront_parameters['distance'] = distance
        if gapH is not None: self.__source_wavefront_parameters['gapH'] = gapH
        if gapV is not None: self.__source_wavefront_parameters['gapV'] = gapV
        if photon_energy is not None: self.__source_wavefront_parameters['photon_energy'] = photon_energy
        if h_slit_points is not None: self.__source_wavefront_parameters['h_slit_points'] = h_slit_points
        if v_slit_points is not None: self.__source_wavefront_parameters['v_slit_points'] = v_slit_points
        if flag_send_wavefront_dimension is not None:
            self.__source_wavefront_parameters['flag_send_wavefront_dimension'] = flag_send_wavefront_dimension


    def get_source_wavefront_parameters(self):
        return self.__source_wavefront_parameters

    def get_dimension(self):  # overwrite this method to export 1D wavefronts if wanted

        if self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 0:
            return 2
        elif self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 1:
            return 1
        elif self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 2:
            return 1

    # from Wofry Decorator
    def get_wavefront(self):
        E, x, y, H, V, I, II = self.calculate_undulator_emission()

        # normalize E
        distance = self.__source_wavefront_parameters['distance']
        # pySRU results are ph/s/0.1bw/mrad^2/distance^2 so I multiply E by sqrt(distance**2)
        wf = GenericWavefront2D.initialize_wavefront_from_arrays(x, y,
                                                                 E[:,:,0].copy() * distance ,
                                                                 z_array_pi=E[:,:,1].copy() * distance,
                                                                 wavelength=1e-10)
        wf.set_photon_energy(self.__source_wavefront_parameters['photon_energy'])

        return wf #, I, x, y, II

    def to_python_code(self, do_plot=True, add_import_section=False):

        txt = ""

        txt += "#"
        txt += "\n# create output_wavefront\n#"
        txt += "\n#"

        txt += "\nfrom orangecontrib.esrf.wofry.util.light_source import WOPySRULightSource # TODO: from wofryimpl..."

        ebeam = self.get_electron_beam()
        und = self.get_magnetic_structure()

        txt += "\nlight_source = WOPySRULightSource.initialize_from_keywords("
        txt += "\n    energy_in_GeV=%g," % ebeam.energy()
        txt += "\n    current=%g," % ebeam.current()
        txt += "\n    K_vertical=%g," % und.K_vertical()
        txt += "\n    period_length=%g," % und.period_length()
        txt += "\n    number_of_periods=%g," % und.number_of_periods()
        txt += "\n    distance=%g," % self.__source_wavefront_parameters['distance']
        txt += "\n    gapH=%g," % self.__source_wavefront_parameters['gapH']
        txt += "\n    gapV=%g," % self.__source_wavefront_parameters['gapV']
        txt += "\n    photon_energy=%g," % self.__source_wavefront_parameters['photon_energy']
        txt += "\n    h_slit_points=%d," % self.__source_wavefront_parameters['h_slit_points']
        txt += "\n    v_slit_points=%d,)" % self.__source_wavefront_parameters['v_slit_points']

        if self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 0:
            txt += "\n\noutput_wavefront = light_source.get_wavefront()"
        elif self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 1:
            txt += "\n\noutput_wavefront = light_source.get_wavefront().get_Wavefront1D_from_profile(0, 0.0)"
        elif self.__source_wavefront_parameters['flag_send_wavefront_dimension'] == 2:
            txt += "\n\noutput_wavefront = light_source.get_wavefront().get_Wavefront1D_from_profile(1, 0.0)"

        return txt

    def calculate_undulator_emission(self):

        from pySRU.Simulation import create_simulation
        from pySRU.ElectronBeam import ElectronBeam as PySruElectronBeam
        from pySRU.MagneticStructureUndulatorPlane import MagneticStructureUndulatorPlane
        from pySRU.TrajectoryFactory import TrajectoryFactory
        from pySRU.RadiationFactory import TRAJECTORY_METHOD_ANALYTIC, TRAJECTORY_METHOD_ODE
        from pySRU.RadiationFactory import RadiationFactory
        from pySRU.RadiationFactory import RADIATION_METHOD_NEAR_FIELD, RADIATION_METHOD_APPROX_FARFIELD

        distance        = self.__source_wavefront_parameters['distance']
        gapH            = self.__source_wavefront_parameters['gapH']
        gapV            = self.__source_wavefront_parameters['gapV']
        photon_energy   = self.__source_wavefront_parameters['photon_energy']
        zero_emittance  = False
        h_slit_points   = self.__source_wavefront_parameters['h_slit_points']
        v_slit_points   = self.__source_wavefront_parameters['v_slit_points']

        print('Running pySRU', distance, gapV, gapH, photon_energy, v_slit_points, h_slit_points)

        hArray = numpy.linspace(-0.5 * gapH, 0.5 * gapH, h_slit_points)
        vArray = numpy.linspace(-0.5 * gapV, 0.5 * gapV, v_slit_points)
        H = numpy.outer(hArray, numpy.ones_like(vArray))
        V = numpy.outer(numpy.ones_like(hArray), vArray)

        myBeam = PySruElectronBeam(Electron_energy = self.get_electron_beam().energy(),
                                   I_current       = self.get_electron_beam().current())

        myUndulator = MagneticStructureUndulatorPlane(K             = self.get_magnetic_structure().K_vertical(),
                                                      period_length = self.get_magnetic_structure().period_length(),
                                                      length        = self.get_magnetic_structure().length())

        simulation_test = create_simulation(magnetic_structure=myUndulator, electron_beam=myBeam,
                                            magnetic_field=None, photon_energy=photon_energy,
                                            traj_method=TRAJECTORY_METHOD_ODE, Nb_pts_trajectory=None,
                                            rad_method=RADIATION_METHOD_NEAR_FIELD, Nb_pts_radiation=None,
                                            initial_condition=None, distance=distance, XY_are_list=False,
                                            X=hArray, Y=vArray)

        # simulation_test.radiation.plot("title=photon energy = %f"%eArray[ie])
        intensArray = simulation_test.radiation.intensity.copy()

        # TODO: this is not nice: I redo the calculations because I need the electric vectors to get wavefront.
        #       this should be avoided after refactoring pySRU to include electric field in simulations!!
        electric_field = simulation_test.radiation_fact.calculate_electrical_field(
            simulation_test.trajectory, simulation_test.source, H, V, distance)

        E = electric_field._electrical_field
        # pol_deg1 = (np.abs(E[:, 0]) / (np.abs(E[:, 0]) + np.abs(E[:, 1]))).flatten()  # SHADOW definition!!



        II = numpy.sum(numpy.abs(E) ** 2, axis=-1)

        # grid in m
        return (E, hArray, vArray, H, V, intensArray, II)



if __name__ == "__main__":


    pp = WOPySRULightSource.initialize_from_keywords(
        name="",
        energy_in_GeV=6.04,
        current=0.2,
        K_vertical=1.68,
        period_length=0.018,
        number_of_periods=222,
        distance=10.0,
        gapH=0.003,
        gapV=0.003,
        photon_energy=7000.0,
        h_slit_points=51,
        v_slit_points=51,
        flag_send_wavefront_dimension=1,
    )

    wf = pp.get_wavefront()

    print(">>>>> Dimension: ", pp.get_dimension())

    from srxraylib.plot.gol import plot_image
    plot_image(wf.get_intensity(), wf.get_coordinate_x(), wf.get_coordinate_y())
    # plot_image(iint, x, y, title="pySRU")
    # plot_image(II, x, y, title="Me")


import numpy, sys

from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.widgets.gui import ConfirmDialog
from oasys.util.oasys_util import EmittingStream, TriggerIn

from syned.widget.widget_decorator import WidgetDecorator
from syned.beamline.element_coordinates import ElementCoordinates
from syned.beamline.beamline_element import BeamlineElement
from syned.beamline.shape import *

from wofry.propagator.propagator import PropagationManager, PropagationElements, PropagationParameters
from wofryimpl.propagator.propagators1D.fresnel import Fresnel1D
from wofryimpl.propagator.propagators1D.fresnel_convolution import FresnelConvolution1D
from wofryimpl.propagator.propagators1D.fraunhofer import Fraunhofer1D
from wofryimpl.propagator.propagators1D.integral import Integral1D
from wofryimpl.propagator.propagators1D.fresnel_zoom import FresnelZoom1D
from wofryimpl.propagator.propagators1D.fresnel_zoom_scaling_theorem import FresnelZoomScaling1D

from orangecontrib.wofry.util.wofry_objects import WofryData
from orangecontrib.wofry.widgets.gui.ow_wofry_widget import WofryWidget

def initialize_default_propagator_1D():
    propagator = PropagationManager.Instance()

    propagator.add_propagator(Fraunhofer1D())
    propagator.add_propagator(Fresnel1D())
    propagator.add_propagator(FresnelConvolution1D())
    propagator.add_propagator(Integral1D())
    propagator.add_propagator(FresnelZoom1D())
    propagator.add_propagator(FresnelZoomScaling1D())

try:
    initialize_default_propagator_1D()
except:
    pass

class OWWOOpticalElement1D(WofryWidget, WidgetDecorator):

    maintainer = "Luca Rebuffi"
    maintainer_email = "luca.rebuffi(@at@)elettra.eu"
    keywords = ["data", "file", "load", "read"]
    category = "Wofry Optical Elements"

    outputs = [{"name":"WofryData",
                "type":WofryData,
                "doc":"WofryData",
                "id":"WofryData"},
               {"name":"Trigger",
                "type": TriggerIn,
                "doc":"Feedback signal to start a new beam simulation",
                "id":"Trigger"},
               ]

    inputs = [("WofryData", WofryData, "set_input"),
              WidgetDecorator.syned_input_data()[0]]

    oe_name         = Setting("")
    p               = Setting(0.0)
    q               = Setting(0.0)
    angle_radial    = Setting(0.0)
    angle_azimuthal = Setting(0.0)

    shape = Setting(0)
    surface_shape = Setting(0)

    input_data = None
    wavefront_to_plot = None

    propagators_list = ["Fresnel", "Fresnel (Convolution)", "Fraunhofer", "Integral", "Fresnel Zoom","Fresnel Zoom Scaled"]
    # plot_titles = ["Wavefront 1D Intensity", "Wavefront 1D Phase","Wavefront Real(Amplitude)","Wavefront Imag(Amplitude)"]

    propagator = Setting(4)
    magnification_x = Setting(1.0) # For Fresnel Zoom & Integral
    magnification_N = Setting(1.0) # For Integral
    scaled_guess_R = Setting(True) # For Fresnel Zoom Scaled
    scaled_R = Setting(1000.0) # For Fresnel Zoom Scaled
    scaled_Rmax = Setting(100.0) # For Fresnel Zoom Scaled
    scaled_N = Setting(100) # For Fresnel Zoom Scaled

    wavefront_radius = 1.0

    def __init__(self,is_automatic=True, show_view_options=True, show_script_tab=True):
        super().__init__(is_automatic=is_automatic, show_view_options=show_view_options, show_script_tab=show_script_tab)

        self.runaction = widget.OWAction("Propagate Wavefront", self)
        self.runaction.triggered.connect(self.propagate_wavefront)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Propagate Wavefront", callback=self.propagate_wavefront)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        gui.separator(self.controlArea)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)


        self.tab_bas = oasysgui.createTabPage(self.tabs_setting, "O.E. Setting")


        oasysgui.lineEdit(self.tab_bas, self, "oe_name", "O.E. Name", labelWidth=260, valueType=str, orientation="horizontal")

        self.coordinates_box = oasysgui.widgetBox(self.tab_bas, "Coordinates", addSpace=True, orientation="vertical")

        tmp = oasysgui.lineEdit(self.coordinates_box, self, "p", "Distance from previous Continuation Plane [m]", labelWidth=280, valueType=float, orientation="horizontal")
        tmp.setToolTip("p")
        tmp = oasysgui.lineEdit(self.coordinates_box, self, "q", "Distance to next Continuation Plane [m]", labelWidth=280, valueType=float, orientation="horizontal")
        tmp.setToolTip("q")
        # Commented srio (not yet implemented) TODO: implement it!
        # oasysgui.lineEdit(self.coordinates_box, self, "angle_radial", "Incident Angle (to normal) [deg]", labelWidth=280, valueType=float, orientation="horizontal")
        # oasysgui.lineEdit(self.coordinates_box, self, "angle_azimuthal", "Rotation along Beam Axis [deg]", labelWidth=280, valueType=float, orientation="horizontal")

        self.draw_specific_box()

        self.create_propagation_setting_tab()


    def create_propagation_setting_tab(self):

        self.tab_pro = oasysgui.createTabPage(self.tabs_setting, "Propagation Setting")
        gui.comboBox(self.tab_pro, self, "propagator", label="Propagator", labelWidth=260,
                     items=self.propagators_list,
                     callback=self.set_Propagator,
                     sendSelectedValue=False, orientation="horizontal")

        # Fresnel
        self.fresnel_box = oasysgui.widgetBox(self.tab_pro, "", addSpace=False, orientation="vertical", height=90)

        # Fraunhoffer
        self.fraunhofer_box = oasysgui.widgetBox(self.tab_pro, "", addSpace=False, orientation="vertical", height=90)

        # Integral
        self.integral_box = oasysgui.widgetBox(self.tab_pro, "", addSpace=False, orientation="vertical", height=90)


        tmp = oasysgui.lineEdit(self.integral_box, self, "magnification_x", "Magnification Factor for interval",
                          labelWidth=260, valueType=float, orientation="horizontal")
        tmp.setToolTip("magnification_x")

        tmp = oasysgui.lineEdit(self.integral_box, self, "magnification_N", "Magnification Factor for N points",
                          labelWidth=260, valueType=float, orientation="horizontal")
        tmp.setToolTip("magnification_N")

        # Fresnel zoom
        self.zoom_box = oasysgui.widgetBox(self.tab_pro, "", addSpace=False, orientation="vertical", height=90)

        tmp = oasysgui.lineEdit(self.zoom_box, self, "magnification_x", "Magnification Factor for interval",
                          labelWidth=260, valueType=float, orientation="horizontal")
        tmp.setToolTip("magnification_x")


        # Fresnel Sacled zoom

        self.zoom_scaled_box = oasysgui.widgetBox(self.tab_pro, "", addSpace=False, orientation="vertical")

        tmp = oasysgui.lineEdit(self.zoom_scaled_box, self, "magnification_x", "Magnification Factor for interval",
                          labelWidth=260, valueType=float, orientation="horizontal")
        tmp.setToolTip("magnification_x")

        gui.comboBox(self.zoom_scaled_box, self, "scaled_guess_R", label="Guess wavefront curvature", labelWidth=260,
                     items=["No","Yes"],
                     callback=self.set_ScaledGuess,
                     sendSelectedValue=False, orientation="horizontal")

        self.zoom_scaled_box_1 = oasysgui.widgetBox(self.zoom_scaled_box, "", addSpace=False, orientation="vertical", height=90)
        self.zoom_scaled_box_2 = oasysgui.widgetBox(self.zoom_scaled_box, "", addSpace=False, orientation="vertical", height=90)

        oasysgui.lineEdit(self.zoom_scaled_box_1, self, "scaled_R", "Wavefront radius of curvature",
                          labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.zoom_scaled_box_2, self, "scaled_Rmax", "Maximum wavefront radius of curvature",
                          labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.zoom_scaled_box_2, self, "scaled_N", "Number of points for guessing curvature",
                          labelWidth=260, valueType=int, orientation="horizontal")

        self.set_Propagator()


    def set_Propagator(self):
        self.fresnel_box.setVisible(self.propagator <= 1)
        self.fraunhofer_box.setVisible(self.propagator == 2)
        self.integral_box.setVisible(self.propagator == 3)
        self.zoom_box.setVisible(self.propagator == 4)
        self.zoom_scaled_box.setVisible(self.propagator == 5)
        if self.propagator == 5: self.set_ScaledGuess()

    def set_ScaledGuess(self):
        self.zoom_scaled_box_1.setVisible(self.scaled_guess_R==0)
        self.zoom_scaled_box_2.setVisible(self.scaled_guess_R==1)

    def draw_specific_box(self):
        pass

    def check_data(self):
        congruence.checkNumber(self.p, "Distance from previous Continuation Plane")
        congruence.checkNumber(self.q, "Distance to next Continuation Plane")
        congruence.checkAngle(self.angle_radial, "Incident Angle (to normal)")
        congruence.checkAngle(self.angle_azimuthal, "Rotation along Beam Axis")

    def propagate_wavefront(self):

        self.progressBarInit()

        self.wofry_output.setText("")

        sys.stdout = EmittingStream(textWritten=self.writeStdOut)

        if self.input_data is None: raise Exception("No Input Data")

        self.check_data()

        # propagation to o.e.

        input_wavefront  = self.input_data.get_wavefront()
        beamline         = self.input_data.get_beamline().duplicate()

        optical_element = self.get_optical_element()
        optical_element.name = self.oe_name if not self.oe_name is None else self.windowTitle()

        beamline_element = BeamlineElement(optical_element=optical_element,
                                           coordinates=ElementCoordinates(p=self.p,
                                                                          q=self.q,
                                                                          angle_radial=numpy.radians(self.angle_radial),
                                                                          angle_azimuthal=numpy.radians(self.angle_azimuthal)))

        #
        # this will store the propagation parameters in beamline in order to perform the propagation in the script
        #
        # 1D
        # ==
        #
        # propagators_list = ["Fresnel",    "Fresnel (Convolution)",  "Fraunhofer",    "Integral",    "Fresnel Zoom",    "Fresnel Zoom Scaled"]
        # class_name       = ["Fresnel1D",  "FresnelConvolution1D",   "Fraunhofer1D",  "Integral1D",  "FresnelZoom1D",   "FresnelZoomScaling1D"]
        # handler_name     = ["FRESNEL_1D", "FRESNEL_CONVOLUTION_1D", "FRAUNHOFER_1D", "INTEGRAL_1D", "FRESNEL_ZOOM_1D", "FRESNEL_ZOOM_SCALING_1D"]

        if self.propagator == 0:
            propagator_info = {
                "propagator_class_name": "Fresnel",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": [],
                "propagator_additional_parameters_values": []}
        elif self.propagator == 1:
            propagator_info = {
                "propagator_class_name": "FresnelConvolution1D",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": [],
                "propagator_additional_parameters_values": []}
        elif self.propagator == 2:
            propagator_info = {
                "propagator_class_name": "Fraunhofer1D",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": [],
                "propagator_additional_parameters_values": []}
        elif self.propagator == 3:
            propagator_info = {
                "propagator_class_name": "Integral1D",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": ['magnification_x', 'magnification_N'],
                "propagator_additional_parameters_values": [self.magnification_x, self.magnification_N]}
        elif self.propagator == 4:
            propagator_info = {
                "propagator_class_name": "FresnelZoom1D",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": ['magnification_x'],
                "propagator_additional_parameters_values": [self.magnification_x]}
        elif self.propagator == 5:
            propagator_info = {
                "propagator_class_name": "FresnelZoomScaling1D",
                "propagator_handler_name": self.get_handler_name(),
                "propagator_additional_parameters_names": ['magnification_x','radius'],
                "propagator_additional_parameters_values": [self.magnification_x, self.wavefront_radius]}

        beamline.append_beamline_element(beamline_element, propagator_info)

        propagation_elements = PropagationElements()
        propagation_elements.add_beamline_element(beamline_element)

        propagation_parameters = PropagationParameters(wavefront=input_wavefront.duplicate(),
                                                       propagation_elements=propagation_elements)

        self.set_additional_parameters(propagation_parameters)

        self.setStatusMessage("Begin Propagation")

        propagator = PropagationManager.Instance()

        output_wavefront = propagator.do_propagation(propagation_parameters=propagation_parameters,
                                                     handler_name=self.get_handler_name())

        self.setStatusMessage("Propagation Completed")

        self.wavefront_to_plot = output_wavefront


        if self.view_type > 0:
            self.initializeTabs()
            self.do_plot_results()
        else:
            self.progressBarFinished()

        self.send("WofryData", WofryData(beamline=beamline, wavefront=output_wavefront))
        self.send("Trigger", TriggerIn(new_object=True))



        # try:
        if True:
            self.wofry_python_script.set_code(beamline.to_python_code())
        # except:
        #     pass

        self.setStatusMessage("")
        try:
            self.print_intensities()
        except:
            pass

    def print_intensities(self):
        input_wavefront = self.input_data.get_wavefront()
        output_wavefront = self.wavefront_to_plot

        c1 = input_wavefront.get_intensity().sum()
        c2 = output_wavefront.get_intensity().sum()
        d1 = input_wavefront.delta()
        d2 = output_wavefront.delta()
        i1 = c1 * d1
        i2 = c2 * d2
        print("\n\n\n ==========  integrated intensities: ")
        print("input wavefront integrated intensity: %g, counts: %g" % (i1, c1))
        print("output wavefront integrated intensity: %g, counts: %g" % (i2, c2))
        print("output/input intensity ratio (transmission): %g " % (i2 / i1))
        print("(input-output)/input intensity ratio (absorption): %g " % ((i1 - i2) / i1))
        print("abscissas step in: %g um, out: %g um" % (1e6 * d1, 1e6 * d2))

    def get_handler_name(self):
        if self.propagator == 0:
            return Fresnel1D.HANDLER_NAME
        elif self.propagator == 1:
            return FresnelConvolution1D.HANDLER_NAME
        elif self.propagator == 2:
            return Fraunhofer1D.HANDLER_NAME
        elif self.propagator == 3:
            return Integral1D.HANDLER_NAME
        elif self.propagator == 4:
            return FresnelZoom1D.HANDLER_NAME
        elif self.propagator == 5:
            return FresnelZoomScaling1D.HANDLER_NAME

    def set_additional_parameters(self, propagation_parameters):
        if self.propagator <= 2:
            pass
        elif self.propagator == 3:
            propagation_parameters.set_additional_parameters("magnification_x", self.magnification_x)
            propagation_parameters.set_additional_parameters("magnification_N", self.magnification_N)
        elif self.propagator == 4:
            propagation_parameters.set_additional_parameters("magnification_x", self.magnification_x)
        elif self.propagator == 5:
            propagation_parameters.set_additional_parameters("magnification_x", self.magnification_x)
            if self.scaled_guess_R:
                # from srxraylib.plot.gol import plot
                # radii,fig_of_mer = self.input_wavefront.scan_wavefront_curvature(
                #     rmin=-self.scaled_Rmax,rmax=self.scaled_Rmax,rpoints=self.scaled_N)
                # plot(radii,fig_of_mer)
                self.wavefront_radius = self.input_data.get_wavefront().guess_wavefront_curvature(
                    rmin=-self.scaled_Rmax,rmax=self.scaled_Rmax,rpoints=self.scaled_N)
                print("Guess wavefront curvature radius: %f m " % self.wavefront_radius)
            else:
                self.wavefront_radius = self.scaled_R
            propagation_parameters.set_additional_parameters("radius", self.wavefront_radius)

    def get_optical_element(self):
        raise NotImplementedError()

    def set_input(self, wofry_data):
        if not wofry_data is None:
            if isinstance(wofry_data, WofryData):
                self.input_data = wofry_data
            else:
                raise Exception("Only wofry_data allowed as input")

            if self.is_automatic_execution:
                self.propagate_wavefront()

    def get_titles(self):
        return ["Wavefront 1D Intensity",
                "Wavefront 1D Phase",
                "Wavefront Real(Amplitude)",
                "Wavefront Imag(Amplitude)"]

    def initializeTabs(self):
        size = len(self.tab)
        indexes = range(0, size)

        for index in indexes:
            self.tabs.removeTab(size-1-index)

        self.tab = []
        self.plot_canvas = []

        for index in range(0, len(self.get_titles())):
            self.tab.append(gui.createTabPage(self.tabs, self.get_titles()[index]))
            self.plot_canvas.append(None)

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

    def do_plot_results(self, progressBarValue=80, closeProgressBar=True):

        if not self.wavefront_to_plot is None:

            self.progressBarSet(progressBarValue)

            self.plot_data1D(x=1e6*self.wavefront_to_plot.get_abscissas(),
                             y=self.wavefront_to_plot.get_intensity(),
                             progressBarValue=progressBarValue,
                             tabs_canvas_index=0,
                             plot_canvas_index=0,
                             title=self.get_titles()[0],
                             xtitle="Spatial Coordinate [$\mu$m]",
                             ytitle="Intensity")


            self.plot_data1D(x=1e6*self.wavefront_to_plot.get_abscissas(),
                             y=self.wavefront_to_plot.get_phase(from_minimum_intensity=0.1,unwrap=1),
                             progressBarValue=progressBarValue + 10,
                             tabs_canvas_index=1,
                             plot_canvas_index=1,
                             title=self.get_titles()[1],
                             xtitle="Spatial Coordinate [$\mu$m]",
                             ytitle="Phase [unwrapped, for intensity > 10% of peak] (rad)")

            self.plot_data1D(x=1e6*self.wavefront_to_plot.get_abscissas(),
                             y=numpy.real(self.wavefront_to_plot.get_complex_amplitude()),
                             progressBarValue=progressBarValue + 10,
                             tabs_canvas_index=2,
                             plot_canvas_index=2,
                             title=self.get_titles()[2],
                             xtitle="Spatial Coordinate [$\mu$m]",
                             ytitle="Real(Amplitude)")

            self.plot_data1D(x=1e6*self.wavefront_to_plot.get_abscissas(),
                             y=numpy.imag(self.wavefront_to_plot.get_complex_amplitude()),
                             progressBarValue=progressBarValue + 10,
                             tabs_canvas_index=3,
                             plot_canvas_index=3,
                             title=self.get_titles()[3],
                             xtitle="Spatial Coordinate [$\mu$m]",
                             ytitle="Imag(Amplitude)")


            # for i in range(len(self.get_titles())):
            #     self.plot_canvas[i].resetZoom()

            if closeProgressBar: self.progressBarFinished()

    def receive_syned_data(self, data):
        if not data is None:
            beamline_element = data.get_beamline_element_at(-1)

            if not beamline_element is None:
                self.oe_name = beamline_element._optical_element._name
                self.p = beamline_element._coordinates._p
                self.q = beamline_element._coordinates._q
                self.angle_azimuthal = round(numpy.degrees(beamline_element._coordinates._angle_azimuthal), 6)
                self.angle_radial = round(numpy.degrees(beamline_element._coordinates._angle_radial), 6)

                self.receive_specific_syned_data(beamline_element._optical_element)
            else:
                raise Exception("Syned Data not correct: Empty Beamline Element")

    def receive_specific_syned_data(self, optical_element):
        raise NotImplementedError()

    def callResetSettings(self):
        if ConfirmDialog.confirmed(parent=self, message="Confirm Reset of the Fields?"):
            try:
                self.resetSettings()
            except:
                pass




#--------------------------------------------------------------

class OWWOOpticalElementWithBoundaryShape1D(OWWOOpticalElement1D):
    # BOUNDARY

    vertical_shift = Setting(0.0)

    height = Setting(0.0)


    def draw_specific_box(self):
        self.shape_box = oasysgui.widgetBox(self.tab_bas, "Boundary Shape", addSpace=True, orientation="vertical")

        # gui.comboBox(self.shape_box, self, "shape", label="Boundary Shape", labelWidth=350,
        #              items=["Rectangle"],
        #              callback=self.set_Shape,
        #              sendSelectedValue=False, orientation="horizontal")

        oasysgui.lineEdit(self.shape_box, self, "vertical_shift", "Shift [m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.rectangle_box = oasysgui.widgetBox(self.shape_box, "", addSpace=False, orientation="vertical", height=60)

        oasysgui.lineEdit(self.rectangle_box, self, "height", "Aperture [m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.circle_box = oasysgui.widgetBox(self.shape_box, "", addSpace=False, orientation="vertical", height=60)

        self.set_Shape()

    def set_Shape(self):
        self.rectangle_box.setVisible(True)
        # self.rectangle_box.setVisible(self.shape == 0)

    def get_boundary_shape(self):
        if self.shape == 0:
            # note that wavefront 1d will be clipped using the first two coordinates!
            boundary_shape = Rectangle(x_left=-0.5*self.height + self.vertical_shift,
                                       x_right=0.5*self.height + self.vertical_shift,
                                       y_bottom=-0.5*self.height + self.vertical_shift,
                                       y_top=0.5*self.height + self.vertical_shift)

        return boundary_shape

    # def get_boundary_shape_python_code(self):
    #     txt = ""
    #
    #     txt += "\nfrom syned.beamline.shape import Rectangle"
    #     txt += "\n# note that wavefront 1d will be clipped using the first two coordinates!"
    #     txt += "\nboundary_shape = Rectangle(x_left=%g,"%(-0.5*self.height + self.vertical_shift)
    #     txt += "x_right=%g,"%(0.5*self.height + self.vertical_shift)
    #     txt += "y_bottom=%g,"%(-0.5*self.height + self.vertical_shift)
    #     txt += "y_top=%g)\n"%(0.5*self.height + self.vertical_shift)
    #
    #     return txt


    def check_data(self):
        super().check_data()

        congruence.checkNumber(self.vertical_shift, "Vertical Shift")
        if self.shape == 0:
            congruence.checkStrictlyPositiveNumber(self.height, "Height")


    def receive_specific_syned_data(self, optical_element):
        if not optical_element is None:
            self.check_syned_instance(optical_element)

            if not optical_element._boundary_shape is None:

                left, right, bottom, top = optical_element._boundary_shape.get_boundaries()

                self.vertical_shift = round(((top + bottom) / 2), 6)

                if isinstance(optical_element._boundary_shape, Rectangle):
                    self.shape = 0

                self.set_Shape()
        else:
            raise Exception("Syned Data not correct: Empty Optical Element")






from orangewidget.settings import Setting
from orangewidget import gui


from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence

from oasys.util.oasys_objects import OasysPreProcessorData, OasysSurfaceData

from syned.widget.widget_decorator import WidgetDecorator

from orangecontrib.esrf.wofry.util.thin_object_corrector import WOThinObjectCorrector #TODO from wofryimpl...

from orangecontrib.wofry.util.wofry_objects import WofryData
from orangecontrib.wofry.widgets.gui.ow_optical_element import OWWOOpticalElement



class OWWOThinObjectCorrector2D(OWWOOpticalElement):

    name = "ThinObjectCorrector2D"
    description = "Wofry: Thin Object Corrector 2D"
    icon = "icons/corrector2d.png"
    priority = 208


    material = Setting(1)
    refraction_index_delta = Setting(5.3e-7)
    att_coefficient = Setting(0.00357382)

    aperture_shape = Setting(0)
    aperture_dimension_v = Setting(100e-6)
    aperture_dimension_h = Setting(200e-6)

    refraction_index_delta = Setting(5.3e-7)
    att_coefficient = Setting(0.00357382)

    correction_method = Setting(1)
    focus_at = Setting(10.0)
    wall_thickness = Setting(0.0)
    apply_correction_to_wavefront = Setting(1)

    write_profile_flag = Setting(0)
    write_profile = Setting("thin_object_profile_2D.h5")

    file_with_thickness_mesh = Setting("profile2D.h5")

    def __init__(self):

        super().__init__(is_automatic=True, show_view_options=True, show_script_tab=True)

    def draw_specific_box(self):

        self.thinobject_box = oasysgui.widgetBox(self.tab_bas, "Thin Object Corrector Setting", addSpace=False, orientation="vertical",
                                           height=350)


        gui.comboBox(self.thinobject_box, self, "correction_method", label="Correction type", labelWidth=350,
                     items=["None","Focus to waist"],
                     callback=self.set_visible,
                     sendSelectedValue=False, orientation="horizontal")

        self.box_corrector_1 = oasysgui.widgetBox(self.thinobject_box, "", addSpace=False, orientation="vertical")

        oasysgui.widgetBox(self.box_corrector_1, "", addSpace=False, orientation="horizontal")
        oasysgui.lineEdit(self.box_corrector_1, self, "focus_at", "Distance to waist [m]",
                          labelWidth=300, valueType=float, orientation="horizontal")


        gui.comboBox(self.box_corrector_1, self, "material", label="Lens material",
                     items=self.get_material_name(),
                     sendSelectedValue=False, orientation="horizontal")

        self.box_refraction_index_id = oasysgui.widgetBox(self.box_corrector_1, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_refraction_index_id, self, "refraction_index_delta", "Refraction index delta",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("refraction_index_delta")

        self.box_att_coefficient_id = oasysgui.widgetBox(self.box_corrector_1, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.box_att_coefficient_id, self, "att_coefficient", "Attenuation coefficient [m-1]",
                          labelWidth=250, valueType=float, orientation="horizontal")
        tmp.setToolTip("att_coefficient")


        self.box_wall_thickness_id = oasysgui.widgetBox(self.box_corrector_1, "", addSpace=False, orientation="horizontal")
        tmp = oasysgui.lineEdit(self.thinobject_box, self, "wall_thickness", "Wall thickness [m]",
                          labelWidth=300, valueType=float, orientation="horizontal", tooltip="wall_thickness")


        oasysgui.lineEdit(self.thinobject_box, self, "file_with_thickness_mesh", "Output file with thickness mesh",
                            labelWidth=200, valueType=str, orientation="horizontal")


        gui.comboBox(self.thinobject_box, self, "apply_correction_to_wavefront", label="Apply correction to wavefront", labelWidth=350,
                     items=["No","Yes"],
                     sendSelectedValue=False, orientation="horizontal")

        self.set_visible()

    def set_visible(self):
        self.box_corrector_1.setVisible(self.correction_method != 0)
        self.box_refraction_index_id.setVisible(self.material in [0])
        self.box_att_coefficient_id.setVisible(self.material in [0])

    def get_material_name(self, index=None):
        materials_list = ["External", "Be", "Al", "Diamond"]
        if index is None:
            return materials_list
        else:
            return materials_list[index]

    def get_optical_element(self):

        return WOThinObjectCorrector(name="Undefined",
                 file_with_thickness_mesh=self.file_with_thickness_mesh,
                 material=self.get_material_name(self.material),
                 refraction_index_delta=self.refraction_index_delta,
                 att_coefficient=self.att_coefficient,
                 correction_method=self.correction_method,
                 focus_at=self.focus_at,
                 wall_thickness=self.wall_thickness,
                 apply_correction_to_wavefront=self.apply_correction_to_wavefront,
                 file_with_thickness_mesh_flag=1,
                                     )

    def check_data(self):
        super().check_data()

    def receive_specific_syned_data(self, optical_element):
        pass

    #
    # overwrite this method to add tab with thickness profile
    #

    def initializeTabs(self):
        size = len(self.tab)
        indexes = range(0, size)

        for index in indexes:
            self.tabs.removeTab(size-1-index)

        titles = ["Intensity","Phase","Thickness Profile"]
        self.tab = []
        self.plot_canvas = []

        for index in range(0, len(titles)):
            self.tab.append(gui.createTabPage(self.tabs, titles[index]))
            self.plot_canvas.append(None)

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)

    def do_plot_results(self, progressBarValue=80):
        super().do_plot_results(progressBarValue)
        if not self.view_type == 0:
            if not self.wavefront_to_plot is None:

                self.progressBarSet(progressBarValue)

                zz, xx, yy = self.get_optical_element().calculate_correction_profile(self.input_data.get_wavefront())

                self.plot_data2D(data2D=zz.T,
                                 dataX=1e6*xx,
                                 dataY=1e6*xx,
                                 progressBarValue=progressBarValue,
                                 tabs_canvas_index=2,
                                 plot_canvas_index=2,
                                 title="thickness profile",
                                 xtitle="Horizontal [$\mu$m] ( %d pixels)"%(xx.size),
                                 ytitle="Vertical [$\mu$m] ( %d pixels)"%(yy.size))

                self.progressBarFinished()
            

if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    def get_example_wofry_data():
        from wofryimpl.propagator.light_source import WOLightSource
        from wofryimpl.beamline.beamline import WOBeamline
        from orangecontrib.wofry.util.wofry_objects import WofryData

        light_source = WOLightSource(dimension=2,
                                     initialize_from=0,
                                     range_from_h=-0.0002,
                                     range_to_h=0.0002,
                                     range_from_v=-0.0002,
                                     range_to_v=0.0002,
                                     number_of_points_h=400,
                                     number_of_points_v=200,
                                     energy=10000.0,
                                     )

        return WofryData(wavefront=light_source.get_wavefront(),
                           beamline=WOBeamline(light_source=light_source))


    a = QApplication(sys.argv)
    ow = OWWOThinObjectCorrector2D()
    ow.set_input(get_example_wofry_data())


    ow.show()
    a.exec_()
    ow.saveSettings()



import numpy
from syned.beamline.optical_elements.mirrors.mirror import Mirror


#
from wofry.beamline.decorators import OpticalElementDecorator

from shadow4.optical_surfaces.s4_conic import S4Conic
from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
from numba import jit, prange

@jit(nopython=True, parallel=True)
def goFromToSequential(field1, x1, y1, x2, y2, wavelength=1e-10, normalize_intensities=False):
    field2 = x2 * 0j
    wavenumber = numpy.pi * 2 / wavelength

    for i in prange(field2.size):
        r = numpy.sqrt(numpy.power(x1 - x2[i], 2) + numpy.power(y1 - y2[i], 2))
        field2[i] = (field1 * numpy.exp(1.j * wavenumber * r)).sum()

    if normalize_intensities:
        field2 *= numpy.sqrt((numpy.abs(field1) ** 2).sum() / (numpy.abs(field2) ** 2).sum())
    return field2

class WOMirror1D(Mirror, OpticalElementDecorator):
    def __init__(self,
                 name="Undefined",
                 surface_shape=None,
                 boundary_shape=None,
                 coating=None,
                 coating_thickness=None,
                 keywords_at_creation=None):

        Mirror.__init__(self, name=name,
                      surface_shape=surface_shape,
                      boundary_shape=boundary_shape,
                      coating=coating, coating_thickness=coating_thickness)

        self._keywords_at_creation = keywords_at_creation


    def get_height_profile(self, input_wavefront):


        shape                          = self._keywords_at_creation["shape"]
        p_focus                        = self._keywords_at_creation["p_focus"]
        q_focus                        = self._keywords_at_creation["q_focus"]
        grazing_angle_in               = self._keywords_at_creation["grazing_angle_in"]
        error_flag                     = self._keywords_at_creation["error_flag"]
        error_file                     = self._keywords_at_creation["error_file"]
        error_file_oversampling_factor = self._keywords_at_creation["error_file_oversampling_factor"]
        mirror_length                  = self._keywords_at_creation["mirror_length"]
        mirror_points                  = self._keywords_at_creation["mirror_points"]


        if error_flag == 0:  # no profile file
            x2_oe = numpy.linspace(-0.5 * mirror_length, 0.5 * mirror_length,
                                   mirror_points)  # x1 / numpy.sin(grazing_angle_in)
            y2_oe = numpy.zeros_like(x2_oe)
        else:
            a = numpy.loadtxt(error_file)
            x2_oe = a[:, 0]
            y2_oe = a[:, 1]

            if error_file_oversampling_factor != 1:
                xnew = numpy.linspace(x2_oe[0], x2_oe[-1], int(x2_oe.size * error_file_oversampling_factor))
                ynew = numpy.interp(xnew, x2_oe, y2_oe)
                x2_oe = xnew
                y2_oe = ynew


        if shape == 0:
            height = numpy.zeros_like(x2_oe)
        elif shape == 1:
            ccc = S4Conic.initialize_as_sphere_from_focal_distances(p_focus, q_focus, grazing_angle_in)
            height = ccc.height(x2_oe)

            print(ccc.info())
            y2_oe += height

        elif shape == 2:
            ccc = S4Conic.initialize_as_ellipsoid_from_focal_distances(p_focus, q_focus, grazing_angle_in)
            height = ccc.height(x2_oe)
            print(ccc.info())
            y2_oe += height
        elif shape == 3:
            ccc = S4Conic.initialize_as_paraboloid_from_focal_distances(p_focus, q_focus, grazing_angle_in)
            height = ccc.height(x2_oe)
            print(ccc.info())
            y2_oe += height
        else:
            raise Exception("Wrong shape")

        return x2_oe, y2_oe


    def get_footprint(self, input_wavefront):

        grazing_angle_in   = self._keywords_at_creation["grazing_angle_in"]
        p_distance         = self._keywords_at_creation["p_distance"]

        # TODO avoid recalculation??
        x2_oe, y2_oe = self.get_height_profile(input_wavefront)

        field2  = self.propagator1D_offaxis_up_to_mirror(input_wavefront, x2_oe, y2_oe,
                                                    p_distance, grazing_angle_in)

        return x2_oe, y2_oe, field2



    def applyOpticalElement(self, input_wavefront, parameters=None, element_index=None):


        grazing_angle_in   = self._keywords_at_creation["grazing_angle_in"]
        flip               = self._keywords_at_creation["flip"]
        p_distance         = self._keywords_at_creation["p_distance"]
        q_distance         = self._keywords_at_creation["q_distance"]
        zoom_factor        = self._keywords_at_creation["zoom_factor"]
        write_profile      = self._keywords_at_creation["write_profile"]

        x2_oe, y2_oe = self.get_height_profile(input_wavefront)

        output_wavefront, x2_oe, y2_oe, field2  = self.propagator1D_offaxis(input_wavefront, x2_oe, y2_oe,
                                                    p_distance, q_distance,
                                                    grazing_angle_in,
                                                    zoom_factor=zoom_factor,
                                                    normalize_intensities=True,
                                                    flip=flip)

        # output files
        if write_profile:
            f = open("reflector_profile1D.dat", "w")
            for i in range(x2_oe.size):
                f.write("%g %g\n" % (x2_oe[i], y2_oe[i]))
            f.close()
            print("File reflector_profile1D.dat written to disk.")

        return output_wavefront


    @classmethod
    def propagator1D_offaxis(cls, input_wavefront, x2_oe, y2_oe, p, q, theta_grazing_in, theta_grazing_out=None,
                             zoom_factor=1.0, normalize_intensities=False, flip=0):

        if theta_grazing_out is None:
            theta_grazing_out = theta_grazing_in

        x1 = input_wavefront.get_abscissas()
        field1 = input_wavefront.get_complex_amplitude()
        wavelength = input_wavefront.get_wavelength()

        if flip == 0:
            x1_oe = -p * numpy.cos(theta_grazing_in) + x1 * numpy.sin(theta_grazing_in)
            y1_oe =  p * numpy.sin(theta_grazing_in) + x1 * numpy.cos(theta_grazing_in)
        else:
            x1_oe =  p * numpy.cos(theta_grazing_in) + x1 * numpy.sin(theta_grazing_in)
            y1_oe =  p * numpy.sin(theta_grazing_in) - x1 * numpy.cos(theta_grazing_in)

        # field2 is the electric field in the mirror
        field2 = goFromToSequential(field1, x1_oe, y1_oe, x2_oe, y2_oe,
                                    wavelength=wavelength, normalize_intensities=normalize_intensities)

        x3 = x1 * zoom_factor

        if flip == 0:
            x3_oe = q * numpy.cos(theta_grazing_out) - x3 * numpy.sin(theta_grazing_out)
            y3_oe = q * numpy.sin(theta_grazing_out) + x3 * numpy.cos(theta_grazing_out)
        else:
            x3_oe = -q * numpy.cos(theta_grazing_out) - x3 * numpy.sin(theta_grazing_out)
            y3_oe =  q * numpy.sin(theta_grazing_out) - x3 * numpy.cos(theta_grazing_out)




        # field3 is the electric field in the image plane
        field3 = goFromToSequential(field2, x2_oe, y2_oe, x3_oe, y3_oe,
                                    wavelength=wavelength, normalize_intensities=normalize_intensities)


        output_wavefront = GenericWavefront1D.initialize_wavefront_from_arrays(x3, field3 / numpy.sqrt(zoom_factor),
                                                                                   wavelength=wavelength)

        return output_wavefront, x2_oe, y2_oe, field2

    @classmethod
    def propagator1D_offaxis_up_to_mirror(cls, input_wavefront, x2_oe, y2_oe, p, theta_grazing_in,
                                          normalize_intensities=False):


        x1 = input_wavefront.get_abscissas()
        field1 = input_wavefront.get_complex_amplitude()
        wavelength = input_wavefront.get_wavelength()

        x1_oe = -p * numpy.cos(theta_grazing_in) + x1 * numpy.sin(theta_grazing_in)
        y1_oe =  p * numpy.sin(theta_grazing_in) + x1 * numpy.cos(theta_grazing_in)

        # field2 is the electric field in the mirror
        field2 = goFromToSequential(field1, x1_oe, y1_oe, x2_oe, y2_oe,
                                    wavelength=wavelength, normalize_intensities=normalize_intensities)

        return field2


    @classmethod
    def create_from_keywords(cls,
                name="mirror 1D",
                shape=0,
                flip=0,
                p_focus=1.0,
                q_focus=1.0,
                grazing_angle_in=0.003,
                p_distance=1.0,
                q_distance=1.0,
                zoom_factor=1.0,
                error_flag=0,
                error_file="",
                error_file_oversampling_factor=1.0,
                mirror_length=1.0,
                mirror_points=100,
                write_profile=0,
                             ):

        keywords_at_creation = {}
        keywords_at_creation["name"]                           = name
        keywords_at_creation["shape"]                          = shape
        keywords_at_creation["flip"]                           = flip
        keywords_at_creation["p_focus"]                        = p_focus
        keywords_at_creation["q_focus"]                        = q_focus
        keywords_at_creation["grazing_angle_in"]               = grazing_angle_in
        keywords_at_creation["p_distance"]                     = p_distance
        keywords_at_creation["q_distance"]                     = q_distance
        keywords_at_creation["zoom_factor"]                    = zoom_factor
        keywords_at_creation["error_flag"]                     = error_flag
        keywords_at_creation["error_file"]                     = error_file
        keywords_at_creation["error_file_oversampling_factor"] = error_file_oversampling_factor
        keywords_at_creation["mirror_length"]                  = mirror_length
        keywords_at_creation["mirror_points"]                  = mirror_points
        keywords_at_creation["write_profile"]                  = write_profile

        out = WOMirror1D(name="Undefined",
                 surface_shape=None,
                 boundary_shape=None,
                 coating=None,
                 coating_thickness=None,
                 keywords_at_creation=keywords_at_creation)

        # out._keywords_at_creation = keywords_at_creation

        return out

    def to_python_code(self, do_plot=False):
        if self._keywords_at_creation is None:
            raise Exception("Python code autogenerated only if created with WOLens.create_from_keywords()")

        txt = "\n"
        txt += "\nfrom orangecontrib.esrf.wofry.util.mirror import WOMirror1D"
        txt += "\n"
        txt += "\noptical_element = WOMirror1D.create_from_keywords("
        txt += "\n    name='%s'," % self._keywords_at_creation["name"]
        txt += "\n    shape=%d," % self._keywords_at_creation["shape"]
        txt += "\n    flip=%d," % self._keywords_at_creation["flip"]
        txt += "\n    p_focus=%g," % self._keywords_at_creation["p_focus"]
        txt += "\n    q_focus=%g," % self._keywords_at_creation["q_focus"]
        txt += "\n    grazing_angle_in=%g," % self._keywords_at_creation["grazing_angle_in"]
        txt += "\n    p_distance=%g," % self._keywords_at_creation["p_distance"]
        txt += "\n    q_distance=%g," % self._keywords_at_creation["q_distance"]
        txt += "\n    zoom_factor=%g," % self._keywords_at_creation["zoom_factor"]
        txt += "\n    error_flag=%d," % self._keywords_at_creation["error_flag"]
        txt += "\n    error_file='%s'," % self._keywords_at_creation["error_file"]
        txt += "\n    error_file_oversampling_factor=%g," % self._keywords_at_creation["error_file_oversampling_factor"]
        txt += "\n    mirror_length=%g," % self._keywords_at_creation["mirror_length"]
        txt += "\n    mirror_points=%d," % self._keywords_at_creation["mirror_points"]
        txt += "\n    write_profile=%d)" % self._keywords_at_creation["write_profile"]
        txt += "\n"
        return txt

#
#
#


if __name__ == "__main__":

    womirror = WOMirror1D.create_from_keywords()
    print(womirror.info())

    print(womirror.to_python_code())
    for key in womirror._keywords_at_creation.keys():
        print(key, womirror._keywords_at_creation[key])


    from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
    input_wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.0005, x_max=0.0005,
                                                                         number_of_points=1000)
    input_wavefront.set_photon_energy(10000)
    input_wavefront.set_spherical_wave(radius=13.73, center=0, complex_amplitude=complex(1, 0))

    output_wavefront = womirror.applyOpticalElement(input_wavefront=input_wavefront)

    from srxraylib.plot.gol import plot

    plot(input_wavefront.get_abscissas(), input_wavefront.get_intensity())
    plot(output_wavefront.get_abscissas(),output_wavefront.get_intensity())






__author__ = 'labx'

import sys
import code
import keyword
import itertools

from PyQt5 import QtGui, QtWidgets

from PyQt5.QtCore import QItemSelectionModel
from PyQt5.QtGui import (
    QTextCursor, QFont, QColor, QPalette
)

from PyQt5.QtCore import Qt, QRegExp


def text_format(foreground=Qt.black, weight=QFont.Normal):
    fmt = QtGui.QTextCharFormat()
    fmt.setForeground(QtGui.QBrush(foreground))
    fmt.setFontWeight(weight)
    return fmt


class PythonSyntaxHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, parent=None):

        self.keywordFormat = text_format(Qt.blue, QFont.Bold)
        self.stringFormat = text_format(Qt.darkGreen)
        self.defFormat = text_format(Qt.black, QFont.Bold)
        self.commentFormat = text_format(Qt.lightGray)
        self.decoratorFormat = text_format(Qt.darkGray)

        self.keywords = list(keyword.kwlist)

        self.rules = [(QRegExp(r"\b%s\b" % kwd), self.keywordFormat)
                      for kwd in self.keywords] + \
                     [(QRegExp(r"\bdef\s+([A-Za-z_]+[A-Za-z0-9_]+)\s*\("),
                       self.defFormat),
                      (QRegExp(r"\bclass\s+([A-Za-z_]+[A-Za-z0-9_]+)\s*\("),
                       self.defFormat),
                      (QRegExp(r"'.*'"), self.stringFormat),
                      (QRegExp(r'".*"'), self.stringFormat),
                      (QRegExp(r"#.*"), self.commentFormat),
                      (QRegExp(r"@[A-Za-z_]+[A-Za-z0-9_]+"),
                       self.decoratorFormat)]

        self.multilineStart = QRegExp(r"(''')|" + r'(""")')
        self.multilineEnd = QRegExp(r"(''')|" + r'(""")')

        super().__init__(parent)

    def highlightBlock(self, text):
        for pattern, format in self.rules:
            exp = QRegExp(pattern)
            index = exp.indexIn(text)
            while index >= 0:
                length = exp.matchedLength()
                if exp.numCaptures() > 0:
                    self.setFormat(exp.pos(1), len(str(exp.cap(1))), format)
                else:
                    self.setFormat(exp.pos(0), len(str(exp.cap(0))), format)
                index = exp.indexIn(text, index + length)

        # Multi line strings
        start = self.multilineStart
        end = self.multilineEnd

        self.setCurrentBlockState(0)
        startIndex, skip = 0, 0
        if self.previousBlockState() != 1:
            startIndex, skip = start.indexIn(text), 3
        while startIndex >= 0:
            endIndex = end.indexIn(text, startIndex + skip)
            if endIndex == -1:
                self.setCurrentBlockState(1)
                commentLen = len(text) - startIndex
            else:
                commentLen = endIndex - startIndex + 3
            self.setFormat(startIndex, commentLen, self.stringFormat)
            startIndex, skip = (start.indexIn(text,
                                              startIndex + commentLen + 3),
                                3)


class PythonScriptEditor(QtWidgets.QPlainTextEdit):
    INDENT = 4

    def lastLine(self):
        text = str(self.toPlainText())
        pos = self.textCursor().position()
        index = text.rfind("\n", 0, pos)
        text = text[index: pos].lstrip("\n")
        return text

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            text = self.lastLine()
            indent = len(text) - len(text.lstrip())
            if text.strip() == "pass" or text.strip().startswith("return "):
                indent = max(0, indent - self.INDENT)
            elif text.strip().endswith(":"):
                indent += self.INDENT
            super().keyPressEvent(event)
            self.insertPlainText(" " * indent)
        elif event.key() == Qt.Key_Tab:
            self.insertPlainText(" " * self.INDENT)
        elif event.key() == Qt.Key_Backspace:
            text = self.lastLine()
            if text and not text.strip():
                cursor = self.textCursor()
                for i in range(min(self.INDENT, len(text))):
                    cursor.deletePreviousChar()
            else:
                super().keyPressEvent(event)

        else:
            super().keyPressEvent(event)


class PythonConsole(QtWidgets.QPlainTextEdit, code.InteractiveConsole):
    def __init__(self, locals=None, parent=None):
        QtWidgets.QPlainTextEdit.__init__(self, parent)
        code.InteractiveConsole.__init__(self, locals)
        self.history, self.historyInd = [""], 0
        self.loop = self.interact()
        next(self.loop)
        self.setStyleSheet("background-color:black; color: white; font-family: Courier, monospace;")

    def setLocals(self, locals):
        self.locals = locals


    def flush(self):
        pass

    def interact(self, banner=None):
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "
        cprt = ('Type "help", "copyright", "credits" or "license" '
                'for more information.')
        if banner is None:
            self.write("Python %s on %s\n%s\n(%s)\n" %
                       (sys.version, sys.platform, cprt,
                        self.__class__.__name__))
        else:
            self.write("%s\n" % str(banner))
        more = 0
        while 1:
            try:
                if more:
                    prompt = sys.ps2
                else:
                    prompt = sys.ps1
                self.new_prompt(prompt)
                yield
                try:
                    line = self.raw_input(prompt)
                except EOFError:
                    self.write("\n")
                    break
                else:
                    more = self.push(line)
            except KeyboardInterrupt:
                self.write("\nKeyboardInterrupt\n")
                self.resetbuffer()
                more = 0

    def raw_input(self, prompt):
        input = str(self.document().lastBlock().previous().text())
        return input[len(prompt):]

    def new_prompt(self, prompt):
        self.write(prompt)
        self.newPromptPos = self.textCursor().position()

    def write(self, data):
        cursor = QTextCursor(self.document())
        cursor.movePosition(QTextCursor.End, QTextCursor.MoveAnchor)
        cursor.insertText(data)
        self.setTextCursor(cursor)
        self.ensureCursorVisible()

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def push(self, line):
        if self.history[0] != line:
            self.history.insert(0, line)
        self.historyInd = 0

        saved = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = self, self
            return code.InteractiveConsole.push(self, line)
        finally:
            sys.stdout, sys.stderr = saved

    def setLine(self, line):
        cursor = QTextCursor(self.document())
        cursor.movePosition(QTextCursor.End)
        cursor.setPosition(self.newPromptPos, QTextCursor.KeepAnchor)
        cursor.removeSelectedText()
        cursor.insertText(line)
        self.setTextCursor(cursor)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            self.write("\n")
            next(self.loop)
        elif event.key() == Qt.Key_Up:
            self.historyUp()
        elif event.key() == Qt.Key_Down:
            self.historyDown()
        elif event.key() == Qt.Key_Tab:
            self.complete()
        elif event.key() in [Qt.Key_Left, Qt.Key_Backspace]:
            if self.textCursor().position() > self.newPromptPos:
                QtWidgets.QPlainTextEdit.keyPressEvent(self, event)
        else:
            QtWidgets.QPlainTextEdit.keyPressEvent(self, event)

    def historyUp(self):
        self.setLine(self.history[self.historyInd])
        self.historyInd = min(self.historyInd + 1, len(self.history) - 1)

    def historyDown(self):
        self.setLine(self.history[self.historyInd])
        self.historyInd = max(self.historyInd - 1, 0)

    def complete(self):
        pass

    def _moveCursorToInputLine(self):
        """
        Move the cursor to the input line if not already there. If the cursor
        if already in the input line (at position greater or equal to
        `newPromptPos`) it is left unchanged, otherwise it is moved at the
        end.

        """
        cursor = self.textCursor()
        pos = cursor.position()
        if pos < self.newPromptPos:
            cursor.movePosition(QTextCursor.End)
            self.setTextCursor(cursor)

    def pasteCode(self, source):
        """
        Paste source code into the console.
        """
        self._moveCursorToInputLine()

        for line in interleave(source.splitlines(), itertools.repeat("\n")):
            if line != "\n":
                self.insertPlainText(line)
            else:
                self.write("\n")
                next(self.loop)

    def insertFromMimeData(self, source):
        """
        Reimplemented from QPlainTextEdit.insertFromMimeData.
        """
        if source.hasText():
            self.pasteCode(str(source.text()))
            return


def interleave(seq1, seq2):
    """
    Interleave elements of `seq2` between consecutive elements of `seq1`.

        >>> list(interleave([1, 3, 5], [2, 4]))
        [1, 2, 3, 4, 5]

    """
    iterator1, iterator2 = iter(seq1), iter(seq2)
    leading = next(iterator1)
    for element in iterator1:
        yield leading
        yield next(iterator2)
        leading = element

    yield leading


class Script(object):
    Modified = 1
    MissingFromFilesystem = 2

    def __init__(self, name, script, flags=0, filename=None):
        self.name = name
        self.script = script
        self.flags = flags
        self.filename = filename


class ScriptItemDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, parent):
        super().__init__(parent)

    def displayText(self, script, locale):
        if script.flags & Script.Modified:
            return "*" + script.name
        else:
            return script.name

    def paint(self, painter, option, index):
        script = index.data(Qt.DisplayRole)

        if script.flags & Script.Modified:
            option = QtWidgets.QStyleOptionViewItem(option)
            option.palette.setColor(QPalette.Text, QColor(Qt.red))
            option.palette.setColor(QPalette.Highlight, QColor(Qt.darkRed))
        super().paint(painter, option, index)

    def createEditor(self, parent, option, index):
        return QtWidgets.QLineEdit(parent)

    def setEditorData(self, editor, index):
        script = index.data(Qt.DisplayRole)
        editor.setText(script.name)

    def setModelData(self, editor, model, index):
        model[index.row()].name = str(editor.text())


def select_row(view, row):
    """
    Select a `row` in an item view
    """
    selmodel = view.selectionModel()
    selmodel.select(view.model().index(row, 0),
                    QItemSelectionModel.ClearAndSelect)




import numpy
from orangecontrib.esrf.syned.util.lens import Lens # TODO: from syned.beamline.optical_elements.lenses.lens import Lens

from syned.beamline.shape import Convexity, Direction
from syned.beamline.shape import SurfaceShape, Plane, Paraboloid, ParabolicCylinder, Sphere, SphericalCylinder
from syned.beamline.shape import BoundaryShape, Rectangle, Circle, Ellipse, MultiplePatch

from wofry.beamline.decorators import OpticalElementDecorator

from barc4ro.projected_thickness import proj_thick_2D_crl, proj_thick_1D_crl
from scipy import interpolate
import scipy.constants as codata
import xraylib

class WOLens(Lens, OpticalElementDecorator):
    def __init__(self,
                 name="Undefined",
                 surface_shape1=None,
                 surface_shape2=None,
                 boundary_shape=None,
                 material="",
                 thickness=0.0,
                 keywords_at_creation=None):
        Lens.__init__(self, name=name,
                      surface_shape1=surface_shape1, surface_shape2=surface_shape2,
                      boundary_shape=boundary_shape, material=material, thickness=thickness)

        self._keywords_at_creation = keywords_at_creation

    def get_refraction_index(self, photon_energy=10000.0):

        wave_length = codata.h * codata.c / codata.e / photon_energy

        if self.get_material() == "External":
            refraction_index_delta = self._keywords_at_creation["refraction_index_delta"]
            att_coefficient = self._keywords_at_creation["att_coefficient"]
            if self._keywords_at_creation["verbose"]:
                print("\n\n\nRefracion index delta = %g " % (refraction_index_delta))
                print("Attenuation coeff mu = %g m^-1" % (att_coefficient))
            return refraction_index_delta, att_coefficient

        if self.get_material() == "Be": # Be
            element = "Be"
            density = xraylib.ElementDensity(4)
        elif self.get_material() == "Al": # Al
            element = "Al"
            density = xraylib.ElementDensity(13)
        elif self.get_material() == "Diamond": # Diamond
            element = "C"
            density = 3.51
        else:
            raise Exception("Bad material: " + self.get_material())

        refraction_index = xraylib.Refractive_Index(element, photon_energy/1000, density)
        refraction_index_delta = 1 - refraction_index.real
        att_coefficient = 4*numpy.pi * (xraylib.Refractive_Index(element, photon_energy/1000, density)).imag / wave_length

        if self._keywords_at_creation["verbose"]:
            print("\n\n\n ==========  parameters recovered from xraylib : ")
            print("Element: %s" % element)
            print("        density = %g " % density)
            print("Photon energy = %g eV" % (photon_energy))
            print("Refracion index delta = %g " % (refraction_index_delta))
            print("Attenuation coeff mu = %g m^-1" % (att_coefficient))

        return refraction_index_delta, att_coefficient


    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):

        x, y, lens_thickness = self.get_surface_thickness_mesh(wavefront)

        photon_energy = wavefront.get_photon_energy()



        refraction_index_delta, att_coefficient = \
            self.get_refraction_index(photon_energy=photon_energy)

        # this is for info...
        #
        if self._keywords_at_creation["verbose"]:
            print("\n\n\n ==========  parameters in use : ")
            number_of_curved_surfaces = self._keywords_at_creation["number_of_curved_surfaces"]
            lens_radius = self._keywords_at_creation["lens_radius"]
            n_lenses = self._keywords_at_creation["n_lenses"]

            print("\n\nRadius of curvature R = %g um" % (1e6 * lens_radius))
            print("Number of lenses N: %d" % n_lenses)
            print("Number of curved refractive surfaces in a lens Nd = %d" % (number_of_curved_surfaces))
            if number_of_curved_surfaces != 0:
                F = lens_radius / (number_of_curved_surfaces * n_lenses * refraction_index_delta)
                print("Focal distance F = R / (Nd N delta) = %g m" % (F))
        # end info...


        amp_factors = numpy.exp(-1.0 * att_coefficient * lens_thickness / 2) # factor of 2 because it is amplitude
        phase_shifts = -1.0 * wavefront.get_wavenumber() * refraction_index_delta * lens_thickness

        output_wavefront = wavefront.duplicate()
        output_wavefront.rescale_amplitudes(amp_factors)
        output_wavefront.add_phase_shifts(phase_shifts)

        return output_wavefront


    def get_surface_thickness_mesh(self, wavefront):

        _foc_plane, _shape, _apert_h, _apert_v, _r_min, _n, _wall_thickness, _aperture= self.__get_barc_inputs()
        _axis_x = wavefront.get_coordinate_x()
        _axis_y = wavefront.get_coordinate_y()

        if self._keywords_at_creation["verbose"]:
            print("\n\n\n ==========  parameters recovered for barc4ro.proj_thick_2D_crl : ")
            print(">>> _aperture = ", _aperture)
            print(">>> _apert_h = ", _apert_h)
            print(">>> _apert_v = ", _apert_v)
            print(">>> _wall_thick:  min. wall thickness between 'holes' [m]= ", _wall_thickness)
            print(">>> _n: number of lenses (/holes) = ", _n)
            print(">>> _r_min: radius (on tip of parabola for parabolic shape) [m] = ", _r_min)
            print(">>> _shape: 1- parabolic, 2- circular (spherical) = ", _shape)
            print(">>> _foc_plane: plane of focusing: 1- horizontal, 2- vertical, 3- both = ", _foc_plane)
            print(">>> _axis_x : from, to, n = ", _axis_x.min(), _axis_x.max(), _axis_x.size)
            print(">>> _axis_y : from, to, n = ", _axis_y.min(), _axis_y.max(), _axis_y.size)


        x, y, lens_thickness = proj_thick_2D_crl(_foc_plane, _shape, _apert_h, _apert_v, _r_min, _n,
                     _wall_thick=_wall_thickness, _aperture=_aperture,
                     _nx=_axis_x.size, _ny=_axis_y.size,
                     _axis_x=_axis_x, _axis_y=_axis_y,
                     _xc=0, _yc=0,
                     _ang_rot_ex=0, _ang_rot_ey=0, _ang_rot_ez=0,
                     _offst_ffs_x=0, _offst_ffs_y=0,
                     _tilt_ffs_x=0, _tilt_ffs_y=0, _ang_rot_ez_ffs=0,
                     _wt_offst_ffs=0, _offst_bfs_x=0, _offst_bfs_y=0,
                     _tilt_bfs_x=0, _tilt_bfs_y=0, _ang_rot_ez_bfs=0, _wt_offst_bfs=0,
                     isdgr=False, project=True,)

        lens_thickness *= self._keywords_at_creation["n_lenses"]

        return x, y, lens_thickness

    def __get_barc_inputs(self):

    #     :param _foc_plane: plane of focusing: 1- horizontal, 2- vertical, 3- both
        if isinstance(self.get_surface_shape(index=0), ParabolicCylinder) or \
                isinstance(self.get_surface_shape(index=0), SphericalCylinder):
            if self.get_surface_shape(index=0).get_cylinder_direction() == Direction.TANGENTIAL:
                _foc_plane = 2
            elif self.get_surface_shape(index=0).get_cylinder_direction() == Direction.SAGITTAL:
                _foc_plane = 1
            else:
                raise Exception("Wrong _foc_plane value.")
        elif isinstance(self.get_surface_shape(index=0), Paraboloid) or \
                isinstance(self.get_surface_shape(index=0), Sphere) or \
                isinstance(self.get_surface_shape(index=0), Plane):
            _foc_plane = 3
        else:
            raise Exception("Not implemented surface shape")

        if isinstance(self.get_surface_shape(index=0), Paraboloid):
            _shape = 1
        elif isinstance(self.get_surface_shape(index=0), Plane):  # for the moment treated as large parabola
            _shape = 1
        elif isinstance(self.get_surface_shape(index=0), Sphere):
            _shape = 2
        else:
            raise Exception("Wrong _shape value")

        boundaries = self._boundary_shape.get_boundaries()

        #  So, in case of the 2D lens, the aperture can be rectangular with _apert_h and _apert_h,
        #  the case of a circular aperutre, both values must be given, but only _apert_h is considered.

        #     :param _aperture: specifies the type of aperture: circular or square
        #     :param _apert_h: horizontal aperture size [m]
        #     :param _apert_v: vertical aperture size [m]
        if isinstance(self._boundary_shape, Rectangle):
            _aperture = "r"
            _apert_h = boundaries[1] - boundaries[0]
            _apert_v = boundaries[3] - boundaries[2]
        elif isinstance(self._boundary_shape, Circle):
            _aperture = "c"
            _apert_h = 2 * boundaries[0]
            _apert_v = 2 * boundaries[0]
        elif isinstance(self._boundary_shape, Ellipse):
            _aperture = "c"
            _apert_h = 2 * (boundaries[1] - boundaries[0])
            _apert_v = 2 * (boundaries[3] - boundaries[2])  # not used by the library
        else:
            raise NotImplementedError("to be implemented")


        if isinstance(self.get_surface_shape(index=0), Paraboloid):
            _r_min = self.get_surface_shape(index=0).get_parabola_parameter()
        elif isinstance(self.get_surface_shape(index=0), Plane):
            _r_min = 1e18
        elif isinstance(self.get_surface_shape(index=0), Sphere):
            _r_min = self.get_surface_shape(index=0).get_radius()
        else:
            raise NotImplementedError()

        if isinstance(self.get_surface_shape(index=1), Plane):
            _n = 1
        else:
            _n = 2

        _wall_thickness = self.get_thickness()

        return _foc_plane, _shape, _apert_h, _apert_v, _r_min, _n, _wall_thickness, _aperture

    @classmethod
    def create_from_keywords(cls,
                             name="Real Lens",
                             number_of_curved_surfaces=2,
                             two_d_lens=0,
                             surface_shape=0,
                             wall_thickness=10e-6,
                             material="Be",
                             refraction_index_delta=5.3e-07,
                             att_coefficient=0.00357382,
                             lens_radius=100e-6,
                             n_lenses=1,
                             aperture_shape=0,
                             aperture_dimension_h=500e-6,
                             aperture_dimension_v=1000e-6,
                             verbose=1,
                             ):
        if number_of_curved_surfaces == 0:
            surface_shape1 = Plane()
        else:
            #     :param _foc_plane: plane of focusing: 1- horizontal, 2- vertical, 3- both
            if surface_shape == 0:
                if two_d_lens == 0:
                    surface_shape1 = Paraboloid(parabola_parameter=lens_radius, convexity=Convexity.DOWNWARD)
                elif two_d_lens == 1:
                    surface_shape1 = ParabolicCylinder(parabola_parameter=lens_radius, cylinder_direction=Direction.TANGENTIAL, convexity=Convexity.DOWNWARD)
                elif two_d_lens == 2:
                    surface_shape1 = ParabolicCylinder(parabola_parameter=lens_radius, cylinder_direction=Direction.SAGITTAL, convexity=Convexity.UPWARD)
            elif surface_shape == 1:
                if two_d_lens == 0:
                    surface_shape1 = Sphere(radius=lens_radius, convexity=Convexity.DOWNWARD)
                elif two_d_lens == 1:
                    surface_shape1 = SphericalCylinder(radius=lens_radius, cylinder_direction=Direction.TANGENTIAL, convexity=Convexity.DOWNWARD)
                elif two_d_lens == 2:
                    surface_shape1 = SphericalCylinder(radius=lens_radius, cylinder_direction=Direction.SAGITTAL, convexity=Convexity.UPWARD)


        if number_of_curved_surfaces == 0:
            surface_shape2 = Plane()
        elif number_of_curved_surfaces == 1:
            surface_shape2 = Plane()
        elif number_of_curved_surfaces == 2:
            surface_shape2 = surface_shape1   #   not used!

        if aperture_shape == 0:
            boundary_shape = Circle(radius=0.5*aperture_dimension_v)
        elif aperture_shape == 1:
            boundary_shape = Rectangle(x_left=-0.5*aperture_dimension_h, x_right=0.5*aperture_dimension_h,
                                       y_bottom=-0.5*aperture_dimension_v, y_top=0.5*aperture_dimension_v)



        keywords_at_creation = {}
        keywords_at_creation["name"]                          = name
        keywords_at_creation["number_of_curved_surfaces"]     = number_of_curved_surfaces
        keywords_at_creation["two_d_lens"]                    = two_d_lens
        keywords_at_creation["surface_shape"]                 = surface_shape
        keywords_at_creation["wall_thickness"]                = wall_thickness
        keywords_at_creation["material"]                      = material
        keywords_at_creation["refraction_index_delta"]        = refraction_index_delta
        keywords_at_creation["att_coefficient"]               = att_coefficient
        keywords_at_creation["lens_radius"]                   = lens_radius
        keywords_at_creation["n_lenses"]                      = n_lenses
        keywords_at_creation["aperture_shape"]                = aperture_shape
        keywords_at_creation["aperture_dimension_h"]          = aperture_dimension_h
        keywords_at_creation["aperture_dimension_v"]          = aperture_dimension_v
        keywords_at_creation["verbose"]                       = verbose

        out = WOLens(name=name,
                      surface_shape1=surface_shape1,
                      surface_shape2=surface_shape2,
                      boundary_shape=boundary_shape,
                      thickness=wall_thickness,
                      material=material )

        out._keywords_at_creation = keywords_at_creation

        return out

    def to_python_code(self, do_plot=False):
        if self._keywords_at_creation is None:
            raise Exception("Python code autogenerated only if created with WOLens.create_from_keywords()")

        txt = ""
        txt += "\nfrom orangecontrib.esrf.wofry.util.lens import WOLens"
        txt += "\n"
        txt += "\noptical_element = WOLens.create_from_keywords("
        txt += "\n    name='%s',"                    % self._keywords_at_creation["name"]
        txt += "\n    number_of_curved_surfaces=%d," % self._keywords_at_creation["number_of_curved_surfaces"]
        txt += "\n    two_d_lens=%d,"                % self._keywords_at_creation["two_d_lens"]
        txt += "\n    surface_shape=%d,"             % self._keywords_at_creation["surface_shape"]
        txt += "\n    wall_thickness=%g,"            % self._keywords_at_creation["wall_thickness"]
        txt += "\n    material='%s',"                % self._keywords_at_creation["material"]
        if self._keywords_at_creation["material"] == "External":
            txt += "\n    refraction_index_delta=%g, # used if material='External'" % self._keywords_at_creation["refraction_index_delta"]
            txt += "\n    att_coefficient=%g, # used if material='External'"        % self._keywords_at_creation["att_coefficient"]
        txt += "\n    lens_radius=%g,"               % self._keywords_at_creation["lens_radius"]
        txt += "\n    n_lenses=%g,"                  % self._keywords_at_creation["n_lenses"]
        txt += "\n    aperture_shape=%d,"            % self._keywords_at_creation["aperture_shape"]
        txt += "\n    aperture_dimension_h=%g,"      % self._keywords_at_creation["aperture_dimension_h"]
        txt += "\n    aperture_dimension_v=%g)"      % self._keywords_at_creation["aperture_dimension_v"]
        txt += "\n    verbose=%d)"                   % self._keywords_at_creation["verbose"]
        txt += "\n"
        return txt

#
#
#

class WOLens1D(Lens, OpticalElementDecorator):
    def __init__(self,
                 name="Undefined",
                 surface_shape1=None,
                 surface_shape2=None,
                 boundary_shape=None,
                 material="",
                 thickness=0.0,
                 keywords_at_creation=None):
        WOLens.__init__(self, name=name,
                      surface_shape1=surface_shape1, surface_shape2=surface_shape2,
                      boundary_shape=boundary_shape, material=material, thickness=thickness)

        self._keywords_at_creation = keywords_at_creation


    def get_surface_thickness_mesh(self, input_wavefront):

        shape = self._keywords_at_creation["shape"                         ]
        radius = self._keywords_at_creation["radius"                        ]
        lens_aperture = self._keywords_at_creation["lens_aperture"                 ]
        wall_thickness = self._keywords_at_creation["wall_thickness"                ]
        number_of_curved_surfaces  = self._keywords_at_creation["number_of_curved_surfaces" ]
        n_lenses = self._keywords_at_creation["n_lenses"                      ]
        error_flag = self._keywords_at_creation["error_flag"                    ]
        error_file = self._keywords_at_creation["error_file"                    ]
        error_edge_management = self._keywords_at_creation["error_edge_management"         ]
        write_profile_flag = self._keywords_at_creation["write_profile_flag"]
        write_profile = self._keywords_at_creation["write_profile"                 ]
        if self._keywords_at_creation["mis_flag"]:
            xc = self._keywords_at_creation["xc"                            ]
            ang_rot = self._keywords_at_creation["ang_rot"                       ]
            wt_offset_ffs = self._keywords_at_creation["wt_offset_ffs"                 ]
            offset_ffs = self._keywords_at_creation["offset_ffs"                    ]
            tilt_ffs = self._keywords_at_creation["tilt_ffs"                      ]
            wt_offset_bfs = self._keywords_at_creation["wt_offset_bfs"                 ]
            offset_bfs = self._keywords_at_creation["offset_bfs"                    ]
            tilt_bfs = self._keywords_at_creation["tilt_bfs"                      ]
        else:
            xc            = 0
            ang_rot       = 0
            wt_offset_ffs = 0
            offset_ffs    = 0
            tilt_ffs      = 0
            wt_offset_bfs = 0
            offset_bfs    = 0
            tilt_bfs      = 0

        abscissas = input_wavefront.get_abscissas().copy()
        abscissas_on_lens = abscissas

        n_ref_lens = number_of_curved_surfaces

        if n_ref_lens == 0:
            lens_thickness = numpy.full_like(abscissas_on_lens, wall_thickness)
        else:
            if shape == 0:  # Flat
                lens_thickness = numpy.full_like(abscissas_on_lens, wall_thickness)

            elif shape == 1:  # Parabolic

                # focus_length = radius / (n_lenses * n_ref_lens * refraction_index_delta)

                # Implementation of barc4ro
                x_2, lens_thickness = proj_thick_1D_crl(shape, lens_aperture, radius,
                                                        _n=n_ref_lens,
                                                        _wall_thick=wall_thickness,
                                                        _xc=xc,
                                                        _nx=100,
                                                        _ang_rot_ex=ang_rot,
                                                        _offst_ffs_x=offset_ffs,
                                                        _tilt_ffs_x=tilt_ffs,
                                                        _wt_offst_ffs=wt_offset_ffs,
                                                        _offst_bfs_x=offset_bfs,
                                                        _tilt_bfs_x=tilt_bfs,
                                                        _wt_offst_bfs=wt_offset_bfs,
                                                        isdgr=False,
                                                        project=True,
                                                        _axis=abscissas)

            elif shape == 2:  # Circular
                lens_thickness = n_ref_lens * (
                            numpy.abs(radius) - numpy.sqrt(radius ** 2 - abscissas_on_lens ** 2)) + wall_thickness
                bound = 0.5 * lens_aperture
                if radius < bound: bound = radius
                for i, x in enumerate(abscissas_on_lens):
                    if (x < -bound) or (x > bound):
                        lens_thickness[i] = 0
                for i, x in enumerate(abscissas_on_lens):
                    if (x < -bound) or (x > bound):
                        lens_thickness[i] = lens_thickness.max()

        lens_thickness *= n_lenses

        if error_flag:
            a = numpy.loadtxt(error_file)  # extrapolation
            if error_edge_management == 0:
                finterpolate = interpolate.interp1d(a[:, 0], a[:, 1],
                                                    fill_value="extrapolate")  # fill_value=(0,0),bounds_error=False)
            elif error_edge_management == 1:
                finterpolate = interpolate.interp1d(a[:, 0], a[:, 1], fill_value=(0, 0), bounds_error=False)
            else:  # crop
                raise Exception("Bad value of error_edge_management")
            thickness_interpolated = finterpolate(abscissas_on_lens)
            lens_thickness += thickness_interpolated

        # output files
        if write_profile_flag:
            f = open(write_profile, "w")
            for i in range(lens_thickness.size):
                f.write("%g %g\n" % (abscissas_on_lens[i], lens_thickness[i]))
            f.close()
            print("File %s written to disk." % write_profile)

        return abscissas_on_lens, lens_thickness


    def get_refraction_index(self, photon_energy=10000.0):

        wave_length = codata.h * codata.c / codata.e / photon_energy

        if self.get_material() == "External":
            refraction_index_delta = self._keywords_at_creation["refraction_index_delta"]
            att_coefficient = self._keywords_at_creation["att_coefficient"]
            print("\n\n\nRefracion index delta = %g " % (refraction_index_delta))
            print("Attenuation coeff mu = %g m^-1" % (att_coefficient))
            return refraction_index_delta, att_coefficient

        if self.get_material() == "Be": # Be
            element = "Be"
            density = xraylib.ElementDensity(4)
        elif self.get_material() == "Al": # Al
            element = "Al"
            density = xraylib.ElementDensity(13)
        elif self.get_material() == "Diamond": # Diamond
            element = "C"
            density = 3.51
        else:
            raise Exception("Bad material: " + self.get_material())

        refraction_index = xraylib.Refractive_Index(element, photon_energy/1000, density)
        refraction_index_delta = 1 - refraction_index.real
        att_coefficient = 4*numpy.pi * (xraylib.Refractive_Index(element, photon_energy/1000, density)).imag / wave_length

        if self._keywords_at_creation["verbose"]:
            print("\n\n\n ==========  parameters recovered from xraylib : ")
            print("Element: %s" % element)
            print("        density = %g " % density)
            print("Photon energy = %g eV" % (photon_energy))
            print("Refracion index delta = %g " % (refraction_index_delta))
            print("Attenuation coeff mu = %g m^-1" % (att_coefficient))

        return refraction_index_delta, att_coefficient


    def applyOpticalElement(self, input_wavefront, parameters=None, element_index=None):



        refraction_index_delta, att_coefficient = \
            self.get_refraction_index(input_wavefront.get_photon_energy())

        # this is for info...
        if self._keywords_at_creation["verbose"]:
            #
            print("\n\n\n ==========  parameters in use : ")

            radius = self._keywords_at_creation["radius"]
            number_of_curved_surfaces  = self._keywords_at_creation["number_of_curved_surfaces" ]

            n_lenses = self._keywords_at_creation["n_lenses"]

            if number_of_curved_surfaces > 0:
                F = radius / (number_of_curved_surfaces * n_lenses * refraction_index_delta)

                print("\n\nRadius of curvature R = %g um" % (1e6 * radius))
                print("Number of lenses N: %d" % n_lenses)
                print("Number of curved refractive surfaces in a lens Nd = %d" % (number_of_curved_surfaces))
                print("Focal distance F = R / (Nd N delta) = %g m" % (F))
            # end info.



        error_flag = self._keywords_at_creation["error_flag"]
        error_file = self._keywords_at_creation["error_file"]
        error_edge_management = self._keywords_at_creation["error_edge_management"]

        output_wavefront = input_wavefront.duplicate()
        abscissas_on_lens, lens_thickness = self.get_surface_thickness_mesh(input_wavefront=input_wavefront)

        amp_factors = numpy.exp(-1.0 * att_coefficient * lens_thickness / 2) # factor of 2 because it is amplitude
        phase_shifts = -1.0 * output_wavefront.get_wavenumber() * refraction_index_delta * lens_thickness

        output_wavefront.rescale_amplitudes(amp_factors)
        output_wavefront.add_phase_shifts(phase_shifts)

        if error_flag:
            a = numpy.loadtxt(error_file)  # extrapolation
            # profile_limits = a[-1, 0] - a[0, 0]
            profile_limits_projected = a[-1, 0] - a[0, 0]
            wavefront_dimension = output_wavefront.get_abscissas()[-1] - output_wavefront.get_abscissas()[0]
            # print("profile deformation dimension: %f m"%(profile_limits))
            print("profile deformation dimension: %f um" % (1e6 * profile_limits_projected))
            print("wavefront window dimension: %f um" % (1e6 * wavefront_dimension))

            if wavefront_dimension <= profile_limits_projected:
                print("Wavefront window inside error profile domain: no action needed")
            else:
                if error_edge_management == 0:
                    print("Profile deformation extrapolated to fit wavefront dimensions")
                else:
                    output_wavefront.clip(a[0, 0], a[-1, 0])
                    print("Wavefront clipped to limits of deformation profile")

        return output_wavefront

    @classmethod
    def create_from_keywords(cls,
                             name                           ="Real Lens 1D",
                             shape                          =1,
                             radius                         =0.0005,
                             lens_aperture                  =0.001,
                             wall_thickness                 =5e-5,
                             material                       ="Be",
                             refraction_index_delta         =5.3e-07,
                             att_coefficient                =0.00357382,
                             number_of_curved_surfaces      =2,
                             n_lenses                       =1,
                             error_flag                     =0,
                             error_file                     ="",
                             error_edge_management          =0,
                             write_profile_flag             =0,
                             write_profile                  ="",
                             mis_flag                       =0,
                             xc                             =0,
                             ang_rot                        =0,
                             wt_offset_ffs                  =0,
                             offset_ffs                     =0,
                             tilt_ffs                       =0,
                             wt_offset_bfs                  =0,
                             offset_bfs                     =0,
                             tilt_bfs                       =0,
                             verbose                        =1,
                             ):


        keywords_at_creation = {}

        # keywords_at_creation["name"                          ] = name
        keywords_at_creation["shape"                         ] = shape
        keywords_at_creation["radius"                        ] = radius
        keywords_at_creation["lens_aperture"                 ] = lens_aperture
        keywords_at_creation["wall_thickness"                ] = wall_thickness
        # keywords_at_creation["material"                      ] = material
        keywords_at_creation["refraction_index_delta"        ] = refraction_index_delta
        keywords_at_creation["att_coefficient"               ] = att_coefficient
        keywords_at_creation["number_of_curved_surfaces" ]     = number_of_curved_surfaces
        keywords_at_creation["n_lenses"                      ] = n_lenses
        keywords_at_creation["error_flag"                    ] = error_flag
        keywords_at_creation["error_file"                    ] = error_file
        keywords_at_creation["error_edge_management"         ] = error_edge_management
        keywords_at_creation["write_profile_flag"            ] = write_profile_flag
        keywords_at_creation["write_profile"                 ] = write_profile
        keywords_at_creation["mis_flag"                      ] = mis_flag
        keywords_at_creation["xc"                            ] = xc
        keywords_at_creation["ang_rot"                       ] = ang_rot
        keywords_at_creation["wt_offset_ffs"                 ] = wt_offset_ffs
        keywords_at_creation["offset_ffs"                    ] = offset_ffs
        keywords_at_creation["tilt_ffs"                      ] = tilt_ffs
        keywords_at_creation["wt_offset_bfs"                 ] = wt_offset_bfs
        keywords_at_creation["offset_bfs"                    ] = offset_bfs
        keywords_at_creation["tilt_bfs"                      ] = tilt_bfs
        keywords_at_creation["verbose"                       ] = verbose


        return WOLens1D(name=name, material=material, thickness=wall_thickness, keywords_at_creation=keywords_at_creation)

    def to_python_code(self):
        if self._keywords_at_creation is None:
            raise Exception("Python code autogenerated only if created with WOLens.create_from_keywords()")

        txt = ""
        txt += "\nfrom orangecontrib.esrf.wofry.util.lens import WOLens1D"
        txt += "\n"
        txt += "\noptical_element = WOLens1D.create_from_keywords("
        txt += "\n    name='%s'," % self.get_name()
        txt += "\n    shape=%d," % self._keywords_at_creation["shape"]
        txt += "\n    radius=%g," % self._keywords_at_creation["radius"]
        txt += "\n    lens_aperture=%g," % self._keywords_at_creation["lens_aperture"]
        txt += "\n    wall_thickness=%g," % self._keywords_at_creation["wall_thickness"]
        txt += "\n    material='%s'," % self.get_material()
        if self.get_material() == "External":
            txt += "\n    refraction_index_delta=%g, # used if material='External'" % self._keywords_at_creation["refraction_index_delta"]
            txt += "\n    att_coefficient=%g, # used if material='External'" % self._keywords_at_creation["att_coefficient"]
        txt += "\n    number_of_curved_surfaces=%d," % self._keywords_at_creation["number_of_curved_surfaces"]
        txt += "\n    n_lenses=%d," % self._keywords_at_creation["n_lenses"]
        txt += "\n    error_flag=%d," % self._keywords_at_creation["error_flag"]
        txt += "\n    error_file='%s'," % self._keywords_at_creation["error_file"]
        txt += "\n    error_edge_management=%d," % self._keywords_at_creation["error_edge_management"]
        txt += "\n    write_profile_flag=%d," % self._keywords_at_creation["write_profile_flag"]
        txt += "\n    write_profile='%s'," % self._keywords_at_creation["write_profile"]
        txt += "\n    mis_flag=%d," % self._keywords_at_creation["mis_flag"]
        txt += "\n    xc=%g," % self._keywords_at_creation["xc"]
        txt += "\n    ang_rot=%g," % self._keywords_at_creation["ang_rot"]
        txt += "\n    wt_offset_ffs=%g," % self._keywords_at_creation["wt_offset_ffs"]
        txt += "\n    offset_ffs=%g," % self._keywords_at_creation["offset_ffs"]
        txt += "\n    tilt_ffs=%g," % self._keywords_at_creation["tilt_ffs"]
        txt += "\n    wt_offset_bfs=%g," % self._keywords_at_creation["wt_offset_bfs"]
        txt += "\n    offset_bfs=%g," % self._keywords_at_creation["offset_bfs"]
        txt += "\n    tilt_bfs=%g," % self._keywords_at_creation["tilt_bfs"]
        txt += "\n    verbose=%d)" % self._keywords_at_creation["verbose"]
        txt += "\n"
        return txt

if __name__ == "__main__":
    pass

    # wolens = WOLens1D.create_from_keywords()
    # print(wolens.info())
    # for key in wolens._keywords_at_creation.keys():
    #     print(key, wolens._keywords_at_creation[key])
    #
    #
    # from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
    # input_wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.0005, x_max=0.0005,
    #                                                                      number_of_points=1000)
    # input_wavefront.set_photon_energy(10000)
    # input_wavefront.set_spherical_wave(radius=13.73, center=0, complex_amplitude=complex(1, 0))
    #
    # output_wavefront = wolens.applyOpticalElement(input_wavefront=input_wavefront)
    #
    # from srxraylib.plot.gol import plot
    #
    # plot(input_wavefront.get_abscissas(), input_wavefront.get_intensity())
    # plot(output_wavefront.get_abscissas(),output_wavefront.get_intensity())



    # wolens2d = WOLens.create_from_keywords(name='Real Lens 2D',
    # number_of_curved_surfaces=2,
    # two_d_lens=2,
    # surface_shape=0,
    # wall_thickness=5e-05,
    # material='Be',
    # lens_radius=0.0006419,
    # n_lenses=1,
    # aperture_shape=1,
    # aperture_dimension_h=0.0002,
    # aperture_dimension_v=0.001)
    #
    # _foc_plane, _shape, _apert_h, _apert_v, _r_min, _n, _wall_thickness, _aperture = wolens2d.get_barc_inputs()
    #
    # print("\n\n\n ==========  parameters recovered for barc4ro.proj_thick_2D_crl : ")
    # print(">>> _aperture = ", _aperture)
    # print(">>> _apert_h = ", _apert_h)
    # print(">>> _apert_v = ", _apert_v)
    # print(">>> _wall_thick:  min. wall thickness between 'holes' [m]= ", _wall_thickness)
    # print(">>> _n: number of lenses (/holes) = ", _n)
    # print(">>> _r_min: radius (on tip of parabola for parabolic shape) [m] = ", _r_min)
    # print(">>> _shape: 1- parabolic, 2- circular (spherical) = ", _shape)
    # print(">>> _foc_plane: (plane of focusing: 1- horizontal, 2- vertical, 3- both) = ", _foc_plane)
